"""Module to bundle common utilities used in the CLI modules."""

import os
import readline
import shutil
import sys
import tempfile
import typing as ty

import rich
import rich_click as click

from armarx_setup import console, PACKAGE_ROOT, RELEASE_CHANNEL_STABLE
from armarx_setup.core import error, logo
from armarx_setup.core.config import config as axii_config
from armarx_setup.core.util import commands
from armarx_setup.core.util.cli import Console
from armarx_setup.core.util.environ_context import EnvironContext

DISABLE_PROMPTS = False
LOGO_PRINTED = False


class RichConsole(Console):

    def __init__(self):
        self.stdout = rich.console.Console()
        self.stderr = rich.console.Console()

    def print(self, *args, **kwargs):
        self.stdout.print(*args, **kwargs)

    def print_raw(self, *args, **kwargs):
        self.print(*args, **self._raw_kwargs(**kwargs))

    def print_error(self, *args, **kwargs):
        self.stderr.print(":heavy_exclamation_mark: ", *args, **kwargs)

    def print_error_raw(self, *args, **kwargs):
        self.print_error(*args, **self._raw_kwargs(**kwargs))

    def print_exception(self):
        self.stdout.print_exception()
        self.stderr.print()

    @staticmethod
    def _raw_kwargs(**kwargs):
        return {
            "crop": False,
            "overflow": "ignore",
            "no_wrap": True,
            "markup": False,
            "highlight": False,
            **kwargs
        }


def decorated_execute(
        cmd: ty.List[str],
        cwd: ty.Optional[str] = None,
        stdin: ty.Optional[ty.Union[str, commands.StdInBuffer]] = None,
        stdout_callback: ty.Optional[commands.PrintCallback] = None,
        stderr_callback: ty.Optional[commands.PrintCallback] = None,
        env: ty.Optional[ty.Dict[str, str]] = None,
):
    from rich.rule import Rule

    from armarx_setup.core.config import config as axii_config
    console: RichConsole = axii_config.console

    command_printed_output = False

    def stdout_callback_overloaded(text):
        nonlocal command_printed_output
        command_printed_output = True

        console.print_raw(text)
        if stdout_callback is not None:
            stdout_callback(text)

    def stderr_callback_overloaded(text):
        nonlocal command_printed_output
        command_printed_output = True

        console.print_error_raw(text)

        if stderr_callback is not None:
            stderr_callback(text)

    exception: ty.Optional[error.CommandFailed] = None

    with EnvironContext(env, expand_environment_variables=True) as env_ctx:
        cmd_str = [f'"{p}"' if " " in p and not '"' in p else p for p in cmd]
        cmd_info = [f"[b]Run command:[/b] {' '.join(cmd_str)}"]
        if cwd is not None:
            cmd_info.append(f"[b]In working directory:[/b] [link=file://{cwd}]{cwd}[/link]")
        if env_ctx is not None:
            cmd_info.append(f"[b]With environment variables:[/b] {env_ctx.print_env_dict()}")

        console.print(*cmd_info, sep="\n")
        console.print(Rule(title="Command output"))

        try:
            stdout = commands.execute(
                cmd=cmd,
                cwd=cwd,
                stdin=stdin,
                stdout_callback=stdout_callback_overloaded,
                stderr_callback=stderr_callback_overloaded,
                env=None)
        except error.CommandFailed as e:
            error.handle_exception()
            exception = e
            stdout = exception.stdout

    if exception is None:
        return_code = 0
    else:
        return_code = exception.return_code

    if not command_printed_output:
        console.print("[i]Command did not produce any output.[/i]")

    color = 'green' if return_code == 0 else 'red'
    console.print(Rule(title=f"Command exited - Return code [{color}]{return_code}[/{color}]"))

    if exception is not None:
        raise exception

    return stdout


def ensure_prompts_enabled():
    """Ensures that prompts are enabled before actually waiting for user input.

    Raises
    ------
    error.PromptsDisabled
        Thrown if prompts were disabled
    """

    if DISABLE_PROMPTS:
        raise error.PromptsDisabled()


def input_confirm(prompt: str, default_option: bool) -> bool:
    """Query the user for a simple yes or no decision.

    Parameters
    ----------
    prompt: str
        Prompt to show to the user.
    default_option: bool
        The default option (if user enters nothing). Well be prepended if not element of options.

    Returns
    -------
    bool
        True if the user chose 'yes', False if the user chose 'no'.

    Raises
    ------
    error.PromptsDisabled
        Thrown if prompts were disabled
    """

    ensure_prompts_enabled()

    default_option_str: str = "y" if default_option else "n"
    return input_option(prompt, options=["y", "n"], default_option=default_option_str) == "y"


def input_edit(prompt: str, prefill: str) -> str:
    """Query the user for free text input with the ability to prefill a default option.

    Parameters
    ----------
    prompt: str
        Prompt to show to the user.
    prefill: str
        Text that will be prefilled for the user to edit.

    Raises
    ------
    error.PromptsDisabled
        Thrown if prompts were disabled.
    error.UserAborted
        Thrown if user aborts the prompt with Ctrl+C.
    """

    ensure_prompts_enabled()

    readline.set_startup_hook(lambda: readline.insert_text(prefill))
    try:
        return input(prompt)
    except KeyboardInterrupt:
        console.print()
        error.handle_exception()
        raise error.UserAborted() from KeyboardInterrupt
    finally:
        readline.set_startup_hook()


def input_option(prompt: str, options: ty.List[str], default_option: str) -> str:
    """Query the user for one of several options.

    Parameters
    ----------
    prompt: str
        Prompt to show to the user.
    options: List[str]
        The options the user should choose from.
    default_option: str
        The default option (if user enters nothing). Well be prepended if not element of options.

    Returns
    -------
    str
        User decision, which is element of options or default_option.

    Raises
    ------
    error.PromptsDisabled
        Thrown if prompts were disabled.
    """

    ensure_prompts_enabled()

    if default_option not in options:
        options = [default_option] + options

    options_str = [
        f"[b]{option.upper()}[/b]" if option == default_option else option for option in options
    ]

    decision = ""
    while decision not in options:
        decision = console.input(f"{prompt} ({'/'.join(options_str)}) > ").lower()
        if decision == "":
            decision = default_option
    return decision


def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
    print_logo()

    click.rich_click.rich_format_help(self, ctx, formatter)


class AxiiClickGroup(click.Group):
    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        format_help(self, ctx, formatter)

    def get_command(self, ctx, cmd_name):
        def edit_distance_is_deletions_only(command, shorthand):
            """Returns true if the edit distance of command to shorthand consists of deletions only.
            """

            def lcs(str1, str2, m, n):
                L = [[0 for _ in range(n + 1)]
                     for _ in range(m + 1)]

                for i in range(m + 1):
                    for j in range(n + 1):
                        if i == 0 or j == 0:
                            L[i][j] = 0
                        elif str1[i - 1] == str2[j - 1]:
                            L[i][j] = L[i - 1][j - 1] + 1
                        else:
                            L[i][j] = max(L[i - 1][j], L[i][j - 1])

                return L[m][n]

            m = len(command)
            n = len(shorthand)
            length = lcs(command, shorthand, m, n)

            deletions = m - length
            insertions = n - length

            return deletions >= 0 and insertions == 0

        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [command for command in self.list_commands(ctx)
                   if edit_distance_is_deletions_only(command, cmd_name)
                   and command.startswith(cmd_name[0])]
        if not matches:
            return None
        if len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(f"Too many matches. Did you mean any of these: {', '.join(sorted(matches))}?")

    def resolve_command(self, ctx, args):
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        # https://github.com/pallets/click/issues/2402
        return cmd.name if cmd else None, cmd, args


class AxiiClickCommand(click.Command):
    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        format_help(self, ctx, formatter)


group_kwargs = dict(cls=AxiiClickGroup)
command_kwargs = dict(cls=AxiiClickCommand)


def ensure_installation():
    """Ensures that Axii was installed properly. Exists with return code 1 and error message if not.
    """

    axii_installed = os.environ.get("_axii_installed", False)
    if not axii_installed:
        axii_exec = os.path.join(PACKAGE_ROOT, "bin", "axii")
        console.print("[red]Error:[/] Axii is not properly installed. Do so by running the "
                      "following command and follow the steps:")
        print_commands([
            f"{axii_exec} self install --$0"
        ])

        console.print("Afterwards, try re-executing the current command:")
        print_commands([
            f"axii {' '.join(sys.argv[1:])}"
        ])

        sys.exit(1)


def print_logo():
    # Avoid printing the logo multiple times.
    global LOGO_PRINTED

    if not axii_config.get("print_logo") or LOGO_PRINTED:
        return

    from armarx_setup.core.git import repo_status

    LOGO_PRINTED = True
    dirty = True
    dev_mode = axii_config.get("dev_mode")
    release_channel = "unknown"
    stable_release_channel = RELEASE_CHANNEL_STABLE
    axii_repo_status = repo_status(PACKAGE_ROOT, fetch=False)
    commits_ahead = 0
    commits_behind = 0
    if axii_repo_status is not None:
        dirty = axii_repo_status.active_branch != stable_release_channel
        release_channel = axii_repo_status.active_branch
        commits_ahead = axii_repo_status.commits_ahead
        commits_behind = axii_repo_status.commits_behind
    alt_mode = not dev_mode and dirty
    console.print(logo.make_logo_in_panel(alt_mode=alt_mode, title="[i]Welcome to Axii[/]"))

    diagnostics = []

    if dev_mode:
        diagnostics.append(f"[b]Developer mode, on '{release_channel}'[/] (ahead "
                           f"{commits_ahead}, behind {commits_behind}).")
    elif dirty:
        diagnostics.append("[b][red]UNSUPPORTED VERSION[/], switch to stable channel with "
                           f"'axii self change-release {stable_release_channel}'[/].")

    if commits_behind > 0:
        diagnostics.append("[yellow]Update available![/] Please update Axii using "
                           "'axii self update'.")

    if len(diagnostics) > 0:
        for line in diagnostics:
            console.print(line, justify="center")
        console.print()


# Local aliases to use names as keyword arguments in bootstrap and still be able to call those
# functions.
ensure_installation_function = ensure_installation
print_logo_function = print_logo


def bootstrap(ensure_installation: bool = True, print_logo: bool = True):
    if ensure_installation:
        ensure_installation_function()
    if print_logo:
        print_logo_function()


def open_file_in_editor(config_file_path: str, kind: str = "", tmp_file_name=""):
    assert os.path.isfile(config_file_path), f"Not a file: {config_file_path}."

    editor = os.environ.get("EDITOR", "editor")

    console.print(f"Edit {kind} file '{config_file_path}' ...")

    # rb as tf.write() requires binary content.
    with open(config_file_path, "rb") as source:
        content = source.read()

    tmp_file_name = tmp_file_name.replace("/", "∕")

    with tempfile.NamedTemporaryFile(prefix="tmpfile: ", suffix=f" - {tmp_file_name}",
                                     delete=False) as tf:
        tf.write(content)
        tf.flush()

        commands.run(f'{editor} "{tf.name}"', quiet=True)
        tf.close()

        console.print(f"Save {kind} at '{config_file_path}' ...")
        shutil.move(tf.name, config_file_path)


def print_list(items: ty.List[str], numbered: bool = False):
    """Prints a list with bullet points.

    Parameters
    ----------
    items: List[str]
        List to print.
    numbered: bool
        Prints the list with numbers instead of bullet points.
    """

    console.print()
    number = 1
    for item in items:
        bullet = number if numbered else "-"
        console.print(f"  {bullet} {item}")
        number += 1
    console.print()


def print_dictionary(dictionary: ty.Dict[str, str]):
    """Prints a dictionary.

    Parameters
    ----------
    dictionary: Dict[str, str]
        Dictionary to print.
    """

    console.print()
    for key, value in dictionary.items():
        console.print(f"  - {key}: [i #777777]{value}[/]")
    console.print()


def print_commands(command_list: ty.List[str]):
    """Print a list of commands.

    This function applies a special formatting to the output so that the output can conveniently
    be copied form the terminal.

    Parameters
    ----------
    command_list: List[str]
        List of commands to print
    """

    rich_print_command_args = {
        "crop": False,
        "no_wrap": True,
        "overflow": "ignore"
    }

    console.print()
    for command in command_list:
        console.print(f"    {command}", **rich_print_command_args)
    console.print()


def make_decorator(options: ty.List[ty.Callable]) -> ty.Callable:
    def decorator(func):
        for option in reversed(options):
            func = option(func)
        return func

    return decorator
