"""Integrates workspace CLI commands to Axii's core."""

import os
import sys
import typing as ty
from typing import Optional

from xdg import xdg_cache_home

from armarx_setup import console
from armarx_setup.cli import common
from armarx_setup.core import error
from armarx_setup.core.config import axii_config


def get_workspace_by_name_or_active(workspace_name: Optional[str] = None) -> "Workspace":
    from armarx_setup.core.workspace import Workspace

    if workspace_name is not None:
        return Workspace.load_workspace_by_name(workspace_name)
    else:
        return Workspace.load_active_workspace()


def add(**kwargs):
    from armarx_setup.core.workspace import Workspace

    ws = Workspace.load_active_workspace()
    ws.add_module(**kwargs)


def remove(**kwargs):
    from armarx_setup.core.workspace import Workspace

    ws = Workspace.load_active_workspace()
    ws.remove_module(**kwargs)


def add_features(**kwargs):
    from armarx_setup.core.workspace import Workspace

    ws = Workspace.load_active_workspace()
    ws.add_module_features(**kwargs)


def remove_features(**kwargs):
    from armarx_setup.core.workspace import Workspace

    ws = Workspace.load_active_workspace()
    ws.remove_module_features(**kwargs)


def create_workspace(
        workspace_path=None,
        workspace_name=None,
        ccache=False,
        workspace_config_path: ty.Optional[str] = None,
        workspace_default_config_factory: ty.Callable[[], "Config"] = None,
) -> "Workspace":
    if workspace_path is None:
        workspace_path = os.path.join(axii_config.get("default_workspaces_path"), "code")

    if workspace_name is None:
        ws = create_workspace_dialog(workspace_path, ccache_enabled=ccache)
    else:
        from armarx_setup.core.workspace import Workspace
        from armarx_setup.core.workspace.config import Config

        if workspace_default_config_factory is None:
            workspace_default_config_factory = lambda: Config.get_defaults(ccache_enabled=ccache)

        if workspace_config_path is not None:
            config = Config.load(workspace_config_path)
        else:
            config = workspace_default_config_factory()

        ws = Workspace(name=workspace_name, path=workspace_path, config=config)

    console.print()
    console.print(f"Create workspace at '{ws.path}' with config: ")
    create_info = ws.create()
    # Reload config to discard any existing comments.
    ws.load_config()
    common.print_dictionary(create_info)
    console.print("Successfully created the workspace. To activate this workspace, run: ")
    common.print_commands([
        f"axii workspace activate {ws.name}"
    ])
    console.print("To make this workspace the default workspace, run the following command and "
                  "restart your terminal: ")
    common.print_commands([
        f"axii config --global default_workspace {ws.name}"
    ])

    return ws


def create_workspace_dialog(path, ccache_enabled: ty.Optional[bool]=None) -> "Workspace":
    from armarx_setup.core.workspace.config import Config
    from armarx_setup.core.workspace.workspace import Workspace

    current_workspace_env_var = os.environ.get(axii_config.workspace_env_var_name, None)
    if current_workspace_env_var:
        path = current_workspace_env_var

    console.print()
    console.print("Specify the workspace.")

    # Query path.
    path = common.input_edit("Workspace directory: ", path) or path
    path = os.path.abspath(os.path.expandvars(os.path.expanduser(path)))

    # Query for the workspace name given the path, allow user to modify.
    name = os.path.basename(os.path.normpath(path))
    name_ok = False
    known_workspaces = axii_config.get_known_workspaces()
    while not name_ok:
        name = common.input_edit("Workspace name: ", name) or name
        if name in known_workspaces:
            console.print(f"The name {name} is already taken and pointing to "
                          f"'{known_workspaces[name]['path']}'.")
            name_ok = common.input_confirm("Overwrite?", default_option=True)
        else:
            name_ok = True

    full_path = os.path.expanduser(os.path.expandvars(path))

    ccache_enabled = ccache_enabled or common.input_confirm(
        "Enabling Ccache in this workspace to speed up compilation is recommended, but it requires "
        "additional disk space.\nEnable Ccache?", default_option=False)

    config = Config.get_defaults(ccache_enabled=ccache_enabled)

    return Workspace(name=name, path=full_path, config=config)


def list_workspaces(**kwargs):
    console.print("Known workspaces:")
    common.print_dictionary(
        {name: info["path"] for name, info in axii_config.get_known_workspaces().items()})


def activate_workspace(workspace_name, cd: bool = True):
    cache_dir = os.path.join(xdg_cache_home(), "axii")
    ccmd_file_name = "cached_command.sh"
    ccmd_file = os.path.join(cache_dir, ccmd_file_name)

    os.makedirs(cache_dir, exist_ok=True)

    known_workspaces = axii_config.get_known_workspaces()
    if workspace_name not in known_workspaces:
        raise error.NoSuchWorkspace(workspace_name=workspace_name)

    path = known_workspaces[workspace_name]["path"]
    with open(f"{ccmd_file}~", "w", encoding="utf-8") as file:
        file.write(f"export ARMARX_WORKSPACE=\"{path}\"\n")
        if cd:
            file.write(f"cd \"{path}\"\n")
    os.rename(f"{ccmd_file}~", ccmd_file)


def deactivate_workspace(**kwargs):
    from armarx_setup.core.workspace import Workspace

    cache_dir = os.path.join(xdg_cache_home(), "axii")
    ccmd_file_name = "cached_command.sh"
    ccmd_file = os.path.join(cache_dir, ccmd_file_name)

    os.makedirs(cache_dir, exist_ok=True)

    # Throws if no workspace is active.
    Workspace.get_active_workspace_dir()

    with open(f"{ccmd_file}~", "w", encoding="utf-8") as file:
        file.write("_axii_env_reset\n")
        file.write("unset ARMARX_WORKSPACE ARMARX_WORKSPACE_NAME\n")
    os.rename(f"{ccmd_file}~", ccmd_file)


def which(**kwargs):
    from armarx_setup.core.workspace import Workspace

    console.print(Workspace.get_active_workspace_dir())


def auto_remove(yes=False):
    from armarx_setup.core.workspace import Workspace
    ws = Workspace.load_active_workspace()

    auto_remove_executor = ws.auto_remove()

    if len(auto_remove_executor.paths_to_remove) > 0:
        console.print("The following directories do not correspond to any module in the workspace:")
        common.print_list(auto_remove_executor.paths_to_remove)
        console.print("These directories may originate from old modules that are not active or "
                      "required anymore.")

        if yes or common.input_confirm("Proceed deleting?", default_option=False):
            auto_remove_executor.execute()
            console.print("Deleted paths.")
        else:
            console.print("Aborted deletion.")
    else:
        console.print("Nothing to do.")


def env(workspace_name: Optional[str] = None, **kwargs):
    ws = get_workspace_by_name_or_active(workspace_name)
    ws.generate_env(**kwargs)


def system(workspace_name: Optional[str] = None, **kwargs):
    ws = get_workspace_by_name_or_active(workspace_name)
    ws.system(**kwargs)


def update(workspace_name: Optional[str] = None, **kwargs):
    ws = get_workspace_by_name_or_active(workspace_name)
    ws.update(**kwargs)


def prepare(workspace_name: Optional[str] = None, **kwargs):
    ws = get_workspace_by_name_or_active(workspace_name)
    ws.prepare(**kwargs)


def build(workspace_name: Optional[str] = None, **kwargs):
    ws = get_workspace_by_name_or_active(workspace_name)
    ws.build(**kwargs)


def install(workspace_name: Optional[str] = None, **kwargs):
    ws = get_workspace_by_name_or_active(workspace_name)
    ws.install(**kwargs)


def conclude(workspace_name: Optional[str] = None, **kwargs):
    ws = get_workspace_by_name_or_active(workspace_name)
    ws.conclude(**kwargs)


def upgrade(workspace_name: Optional[str] = None, **kwargs):
    from rich.table import Table

    from armarx_setup.core.module.common import StepResult, StepResultStatus
    from armarx_setup.core.util.cli import print_full_width_heading
    from armarx_setup.core.util.prefix_color_map import module_to_colored_text as colorize

    ws = get_workspace_by_name_or_active(workspace_name)
    results = ws.upgrade(**kwargs)

    def status_hr(step_result):
        """Converts the given status to a human-readable form."""

        if isinstance(step_result, StepResult):
            status = step_result.status
        else:
            status = step_result

        if status is None:
            return "[grey35]·[/grey35]"
        if status == StepResultStatus.SUCCESS:
            return "[green]:heavy_check_mark:[/green]"
        if status == StepResultStatus.SUCCESS_WITH_WARNING:
            return "[yellow]:warning:[/yellow]"
        if status == StepResultStatus.FAILURE:
            return "[red]:heavy_multiplication_x:[/red]"
        if status == StepResultStatus.SKIPPED:
            return "[grey46]:arrow_right_hook:[/grey46]"
        return ""

    pbi_heading = "P B I"

    key = [
        "E: [b]E[/b]nvironment",
        "S: [b]S[/b]ystem",
        "U: [b]U[/b]pdate",
        f"{pbi_heading}: [b]P[/b]repare [b]B[/b]uild [b]I[/b]nstall",
        "C: [b]C[/b]onclude",
        f"{status_hr(StepResultStatus.SUCCESS)} Step succeeded",
        f"{status_hr(StepResultStatus.SUCCESS_WITH_WARNING)} Step succeeded with warnings",
        f"{status_hr(StepResultStatus.FAILURE)} Step failed",
        f"{status_hr(StepResultStatus.SKIPPED)} Step was skipped",
        f"{status_hr(None)} No step defined"
    ]

    heading_to_step = {"E": "env", "S": "system", "U": "update", pbi_heading: "?", "C": "conclude"}
    step_headings = heading_to_step.keys()

    table = Table()
    table.add_column("Module", justify="left")
    for column_name in step_headings:
        table.add_column(column_name, justify="center")

    def partition_and_filter(results):
        """Partitions the results to first display those modules with only system steps and all
        remaining afterwards.  Modules without any step are filtered out.
        """

        system_only = {}
        other = {}
        for module_name, step_results in results.items():
            steps = [step for step, step_res in step_results.items() if step_res is not None]
            if set(steps) == {"env", "system"}:
                system_only[module_name] = step_results
            elif len(steps) > 0:
                other[module_name] = step_results
            else:
                # Discard module for display if it doesn't have any steps.
                pass
        return {**system_only, **other}

    results = partition_and_filter(results)

    for module_name, step_results in results.items():
        step_res = []
        for step_heading in step_headings:
            if step_heading == pbi_heading:
                prepare_res = status_hr(step_results["prepare"])
                build_res = status_hr(step_results["build"])
                install_res = status_hr(step_results["install"])
                step_res.append(" ".join([prepare_res, build_res, install_res]))
            else:
                step = heading_to_step[step_heading]
                step_res.append(status_hr(step_results[step]))
        table.add_row(colorize(module_name), *step_res)

    print_full_width_heading("Upgrade summary")
    console.print()
    console.print("[i]Key:[/] " + " | ".join(key))
    console.print(table)


def verify(**kwargs):
    def format_deprecation_message(m: "Module") -> str:
        message = f"[i]{m.general.deprecated.message}"
        message += " " if message.endswith(".") else ". "
        message += "[/i]"
        deletion = ""
        if m.general.deprecated.removal is not None:
            deletion = f" and marked for deletion as of {m.general.deprecated.removal}"
        without_replacement = ""
        if m.general.deprecated.superseded_by is None:
            without_replacement = " [b]without replacement[/b]"
        message += f"\n      The module was declared deprecated{deletion}{without_replacement}."
        return message

    from rich.rule import Rule
    from armarx_setup.core.workspace import Workspace
    from armarx_setup.core.util.prefix_color_map import module_to_colored_text

    issue_found = False

    status_text = "Verifying workspace "

    with console.status(status_text + "...") as status:
        status.update(status_text + "(stray modules or unknown paths) ...")

        ws = Workspace.load_active_workspace()
        paths_to_remove = ws.auto_remove().paths_to_remove

        if len(paths_to_remove) > 0:
            issue_found = True
            s = "s" if len(paths_to_remove) > 1 else ""

            console.print(Rule(title="Report: Unknown Paths in Workspace"))
            console.print(f"A total of {len(paths_to_remove)} path{s} inside the workspace cannot "
                          f"be ascribed to Axii.  This can happen if modules are removed or "
                          f"dependencies change. To fix this, run this dedicated command:")
            common.print_commands(["axii workspace auto-remove"])
            console.print("And follow the instructions.")
            console.print(Rule())
            console.print()

        status.update(status_text + "(deprecated modules) ...")

        deprecated_modules = ws.get_deprecated_modules()
        direct_deprecated_modules = deprecated_modules.directly_added

        if len(direct_deprecated_modules) > 0:
            issue_found = True
            plural = len(direct_deprecated_modules) > 1
            s = "s" if plural else ""
            were = "were" if plural else "was"
            are = "are" if plural else "is"
            these = "these" if plural else "this"

            console.print(Rule(title="Report: Deprecated Modules Added to Workspace"))

            modules_list_str = ""
            for module in direct_deprecated_modules:
                description = f"{format_deprecation_message(module)}\n\n"
                description += "      To fix this:\n\n"
                description += f"      axii workspace remove {module.name}\n"
                if module.general.deprecated.superseded_by is not None:
                    for superseded, _ in module.general.deprecated.superseded_by.items():
                        description += f"      axii workspace add {superseded}\n"
                modules_list_str += f"\n    - {module_to_colored_text(module.name)}: {description}"

            console.print(f"The following module{s} {were} added to the workspace but {are} "
                          f"deprecated:")
            console.print(modules_list_str)
            console.print(f"Run the above commands to resolve {these} issue{s}.")
            console.print(Rule())
            console.print()

        any_deprecated_modules = deprecated_modules.transitive

        if len(any_deprecated_modules) > 0:
            issue_found = True
            plural = len(any_deprecated_modules) > 1
            s = "s" if plural else ""
            are = "are" if plural else "is"

            console.print(Rule(title="Report: Deprecated Modules in Workspace"))

            modules_list_str = ""
            for module in any_deprecated_modules:
                modules_list_str += f"    - {module_to_colored_text(module.name)}: " \
                                    f"{format_deprecation_message(module)}\n"

            console.print(f"The following module{s} {are} included in your workspace but {are} "
                          f"deprecated:")
            console.print()
            console.print(modules_list_str)
            console.print("To trace the dependencies, run:")
            common.print_commands(["axii workspace info"])
            console.print("Consult the corresponding module authors for a possible solution.")
            console.print(Rule())
            console.print()

        status.update(status_text + "(Git repositories) ...")

        git_issues = False
        from armarx_setup.core import git
        for report in git.correct_workspace_configuration_issues(ws, default_checkout=False,
                                                                 self=False, dry_run=True):
            if report is not None and len(report.issues) > 0:
                git_issues = True
                break

        if git_issues:
            issue_found = True

            console.print(Rule(title="Report: Git Repository Issues"))

            console.print("Git repositories in your workspace seem to show configuration issues.  "
                          "To fix this, run this dedicated command:")
            common.print_commands(["axii workspace git reconfigure"])
            console.print("And follow the instructions.")
            console.print(Rule())
            console.print()

    if issue_found:
        console.print("Verification completed. See the above reports for details.")
    else:
        console.print("Verification completed. No problems found.")


def clean(**kwargs):
    from armarx_setup.core.workspace import Workspace

    ws = Workspace.load_active_workspace(enforce_os_fingerprint=False)
    ws.clean(**kwargs)
    ws.env_file.remove("AXII_WORKSPACE_OS_FINGERPRINT", "workspace")
    ws.generate_env()


def test(workspace_name: Optional[str] = None, **kwargs):
    ws = get_workspace_by_name_or_active(workspace_name)
    ws.test(**kwargs)


def git_status(**kwargs):
    from armarx_setup.core import git
    from armarx_setup.core.workspace import Workspace

    ws = Workspace.load_active_workspace()
    git.status(ws, **kwargs)


def git_reconfigure(default_checkout: bool = False, yes: bool = False, self: bool = False):
    from armarx_setup.core import git
    from armarx_setup.core.workspace import Workspace
    from armarx_setup.core.util import prefix_color_map
    from rich.progress import Progress

    ws = Workspace.load_active_workspace()
    total = ws.module_graph.number_of_nodes() + (1 if self else 0)
    all_issues = {}

    with Progress(console=console) as progress:
        task = progress.add_task("Assess Git repositories in workspace", total=total)
        for report in git.correct_workspace_configuration_issues(ws,
                                                                 default_checkout=default_checkout,
                                                                 self=self, dry_run=True):
            if report is not None and len(report.issues) > 0:
                all_issues[report.module.name] = report
            progress.update(task, advance=1)

    # Generate report.
    if len(all_issues) > 0:
        console.print()
        if len(all_issues) == 1:
            console.print("An issue with the current workspace was detected.")
        else:
            console.print("Several issues with the current workspace were detected.")
        for module_name, report in all_issues.items():
            module_name_c = prefix_color_map.module_to_colored_text(module_name)
            console.print(f"Configuration issue{'s' if len(report.issues) > 1 else ''} in "
                          f"{module_name_c}:")
            issues_str = [i.to_string_problem() for i in report.issues]
            common.print_list(issues_str)
    else:
        console.print("No configuration issues detected.")
        return

    # Ask user how to proceed.
    if not yes:
        modules = len(all_issues) > 1
        module_or_modules = "all modules" if modules else "module"
        values = modules or len(list(all_issues.values())[0].issues) == 1
        value_or_values = "values" if values else "value"
        prompt = f"Reconfigure {module_or_modules} to expected {value_or_values}?"
        options = ["y", "n"]
        if values:
            options.append("individual")
        decision = common.input_option(prompt, options=options, default_option="y")
        if decision == "n":
            console.print("No configuration corrections were applied.")
            return
        yes = decision == "y"
        console.print()

    # Proceed according to user decision.
    amount_fixed = 0
    for module_name, report in all_issues.items():
        module_name_c = prefix_color_map.module_to_colored_text(module_name)
        for issue in report.issues:
            console.print(f"{module_name_c}, {issue.to_string_correct()}")
            if yes or common.input_confirm("Reconfigure accordingly?", default_option=True):
                git.correct_module_configuration_issue(report.module, report.repo_status,
                                                       issue=issue, dry_run=False)
                amount_fixed += 1

    # Print report.
    if amount_fixed == 0:
        console.print("No configuration corrections were applied.")
    else:
        s = 's' if amount_fixed > 1 else ''
        console.print(f"Applied {amount_fixed} configuration correction{s}.")


def info(workspace_name: Optional[str] = None, **kwargs):
    ws = get_workspace_by_name_or_active(workspace_name)
    ws.info(**kwargs)


def purge(workspace_name: Optional[str] = None, yes: bool = False):
    ws = get_workspace_by_name_or_active(workspace_name)

    console.print(f"About to purge workspace '{ws.name}' located in {ws.path} ...")
    confirm = yes or common.input_confirm("Are you sure you want to continue purging?",
                                          default_option=False)

    if confirm:
        import shutil
        console.print(f"Purging workspace '{ws.name}' ...")
        try:
            shutil.rmtree(ws.path)
        except Exception as e:
            error.handle_exception()
            console.print(f"Failed to purge workspace '{ws.name}'. Reason: {e}")
        else:
            console.print(f"Purged workspace '{ws.name}'.")
            console.print("Cleaning up metadata ...")
            axii_config.remove_known_workspace(name=ws.name)
            console.print("Cleaned up metadata.")

            if ws.path == ws.get_active_workspace_dir():
                deactivate_workspace()


def exec_(**kwargs):
    from armarx_setup.core.workspace import Workspace

    ws = Workspace.load_active_workspace()
    ws.exec(**kwargs)


def list_modules(deps: bool = False, tree: bool = True):
    from armarx_setup import console
    from armarx_setup.core.workspace import Workspace

    ws = Workspace.load_active_workspace()

    if deps:

        names = [m.name for m in ws.get_modules_in_alphabetical_order()]
        label = "Added modules and their dependencies:"

    else:
        names = list(ws.config.modules)
        label = "Added modules:"

    if len(names) > 0:
        if tree:
            from armarx_setup.core.module import introspection
            tree = introspection.get_module_tree_list(names, label=label)
            console.print(tree)
        else:
            console.print(label)
            common.print_list(names)


def no_workspace_active(e, action=None):
    reason = "No workspace is active.  You can activate a workspace with:"
    if action is not None:
        reason = f"Cannot {action} because no workspace is active.  You can activate a workspace " \
                 f"with:"

    console.print(f"{reason}")
    common.print_commands([
        "axii workspace activate <workspace-name>"
    ])

    console.print("To see which workspaces are known to Axii, run:")
    common.print_commands([
        "axii workspace list"
    ])

    console.print("To activate a workspace permanently, run:")
    common.print_commands([
        "axii config --global default_workspace <workspace-name>"
    ])

    console.print("Alternatively, you can create a new workspace using:")
    common.print_commands([
        "axii workspace create"
    ])

    sys.exit(1)


def workspace_already_exists(e):
    console.print("Cannot create workspace.  The supplied path already exists and is not empty.")

    sys.exit(1)


def no_such_workspace(e):
    known_workspaces = axii_config.get_known_workspaces()
    console.print(f"No such workspace '{e.workspace_name}'. Known workspaces are:")
    common.print_dictionary({name: info["path"] for name, info in known_workspaces.items()})

    sys.exit(1)


if __name__ == '__main__':
    # To suppress type annotation warnings.
    from armarx_setup.core.module import Module
    from armarx_setup.core.workspace import Workspace
