"""Provides Axii's configuration."""

import copy
import inspect
import json
import math
import multiprocessing
import os
import typing as ty
import shutil

from xdg import xdg_config_home

from armarx_setup.core import error
from armarx_setup.core.util.commands import execute
from armarx_setup.core.util.cli import Console, StdConsole
from armarx_setup.core.util.commands import run


__all__ = ["axii_config", "AxiiConfig", "Config", "ConfigVariable", "BuiltInConfigVariables",
           "WORKSPACE_ENV_VAR_NAME", "WORKSPACE_NAME_ENV_VAR_NAME"]


WORKSPACE_ENV_VAR_NAME = "ARMARX_WORKSPACE"
WORKSPACE_NAME_ENV_VAR_NAME = "ARMARX_WORKSPACE_NAME"


class ConfigVariable:
    """Models a generic config variable."""

    def __init__(
            self,
            name: str,
            value_type: ty.Optional[type] = None,
            default_value: ty.Optional[ty.Any] = None,
    ):
        assert value_type is not None or default_value is not None, \
            "Must supply a value_type if no default_value specified."

        self.name: str = name
        self.value_type: type = value_type
        self.value: ty.Optional[value_type] = default_value

        if self.value_type is None:
            self.value_type = type(self.value)

    def verify_value(self, value):
        """Verify the value's format.

        Raises
        ------
        error.InvalidConfigurationValue
            If value does not comply with the required format.
        """

    def parse_value(self, value: ty.Any) -> ty.Any:
        """Parses a string value and casts it to the corresponding variable type.

        Parameters
        ----------
        value: str
            Value from user input or other string representation.

        Returns
        -------
        typing.Any
            Parsed and cast value.
        """

        if isinstance(value, self.value_type):
            return value

        if self.value_type is Config:
            return Config().from_dict(**value)

        self.verify_value(value)

        if self.value_type == bool:
            return self._parse_bool(value)
        if self.value_type == int:
            return int(value)
        return value

    def get_value(self, interpret: bool = False) -> ty.Any:
        """Get the variable's value."""

        value = self.value
        if interpret:
            value = self.interpret_value(value)
        return value

    def set_value(self, value: ty.Any, parse: bool = False):
        """Set the variable's value."""

        if parse or self.value_type is Config:
            value = self.parse_value(value)
        self.value = value

    def interpret_value(self, value):
        """Interprets the value and returns it as another type.

        The default implementation is equivalent to parse_value.
        """

        return self.parse_value(value)

    def _parse_bool(self, value: str) -> bool:
        true_values = ["true", "yes", "on", "1", "y", "enable"]
        false_values = ["false", "no", "off", "0", "n", "disable"]

        if value.lower() in true_values:
            return True
        if value.lower() in false_values:
            return False
        raise error.InvalidConfigurationValue(
            variable=self.name,
            value=value,
            description=f"Value must be any of {', '.join(true_values)} for true values "
                        f"or any of {', '.join(false_values)} for false values")


class Config:
    """Models a generic config."""

    def __init__(
            self,
            config_vars: ty.Optional[ty.Dict[str, ConfigVariable]] = None
    ):
        if config_vars is None:
            config_vars = {}
        self.config_vars: ty.Dict[str, ConfigVariable] = config_vars

    @classmethod
    def from_dict(cls, **kwargs) -> "Config":
        """Builds an instance of Config from a dict."""

        config_vars = {}
        for key, value in kwargs.items():
            value_type = Config if isinstance(value, dict) else type(value)
            cfg_var = ConfigVariable(name=key, value_type=value_type)
            cfg_var.set_value(value)
            config_vars[key] = cfg_var
        return cls(config_vars=config_vars)

    def to_dict(self) -> ty.Dict[str, ty.Any]:
        """Serializes this instance into a dict."""

        out_dict = {}

        for name, cfg_var in self.config_vars.items():
            value = cfg_var.value
            if isinstance(value, Config):
                value = value.to_dict()
            out_dict[name] = value

        return out_dict

    def __str__(self):
        return self.to_dict().__str__()

    def has_variable(self, name: str) -> bool:
        return name in self.config_vars

    def get(self, name: str, interpret: bool = True) -> ty.Any:
        """Get a config variable's value."""

        if name in self.config_vars:
            return self.config_vars[name].get_value(interpret=interpret)
        raise error.NoSuchConfigVariable(name)

    def set(self, name: str, value: ty.Any, parse: bool = True):
        """Set a config variable's value."""

        if name not in self.config_vars:
            self.config_vars[name] = ConfigVariable(name=name, value_type=type(value))
        self.config_vars[name].set_value(value, parse=parse)

    def unset(self, name: str):
        """Unset a config variable."""

        if name not in self.config_vars:
            raise error.NoSuchConfigVariable(name)
        del self.config_vars[name]


class Subconfig(Config):
    pass


class SubconfigVariable(ConfigVariable):
    def __init__(self, name: str):
        default_value = Subconfig()
        super().__init__(name=name, default_value=default_value)

    def parse_value(self, value: ty.Dict[str, str]) -> Config:
        return Subconfig.from_dict(**value)

    def interpret_value(self, value):
        return value


class BuiltInConfigVariable(ConfigVariable):
    """Models a built-in config variable."""

    def __init__(
            self,
            name: str,
            default_value: ty.Optional[ty.Any] = None,
            description: ty.Optional[str] = None
    ):
        value_type = type(default_value)
        super().__init__(name=name, value_type=value_type)

        self.default_value: ty.Optional[value_type] = default_value
        self.value: ty.Optional[value_type] = default_value
        self.description = description

    def get_allowed_values(self):
        """Get a list of all allowed values of this config variable."""

        if self.value_type == bool:
            return ["true", "false"]
        return []


class DefaultWorkspaceConfigVariable(BuiltInConfigVariable):
    """Special implementation for the 'default_workspace' configuration variable."""

    def __init__(self):
        name = "default_workspace"
        default_value = ""
        description = "Sets the default workspace to activate when opening a new terminal."
        super().__init__(name=name, default_value=default_value, description=description)

    def get_allowed_values(self):
        from armarx_setup.core.workspace import Workspace
        return Workspace.get_all_workspace_names()

    def verify_value(self, value: str):
        allowed_values = self.get_allowed_values()

        if value not in allowed_values:
            description = f"Value must be any of {', '.join(allowed_values)}."
            raise error.InvalidConfigurationValue(variable=self.name,
                                                  value=value,
                                                  description=description)

        return value


class GitRemotePreferencesSubconfig(Config):
    """Override for a Config specific for git_remote_preferences with method
    'preference_for_host'."""

    def preference_for_host(self, host: str) -> str:
        """Gets the configured preference for given host.  Falls back to 'https' if nothing
        configured.

        Parameters
        ----------
        host: str
            Host to get the remote preference for.

        Returns
        -------
        str
            'ssh' if preference was configured for host, otherwise 'https'.
        """

        if host in self.config_vars:
            return self.config_vars[host].value
        return "https"  # Default is HTTPS as it requires no additional setup.


class GitRemotePreferencesSubconfigVariable(BuiltInConfigVariable):
    """Special implementation for the git_remote_preferences configuration variable."""

    def __init__(self):
        name = "git_remote_preferences"
        default_value = GitRemotePreferencesSubconfig()
        description = ""
        super().__init__(name=name, default_value=default_value, description=description)

    def parse_value(self, value: ty.Dict[str, str]) -> Config:
        return GitRemotePreferencesSubconfig.from_dict(**value)

    def interpret_value(self, value):
        return value


class UpdateJobNumberConfigVariable(BuiltInConfigVariable):
    """Special implementation for the update_job_number configuration variable."""

    def __init__(self):
        name = "update_job_number"
        default_value = 32
        description = "Defines the job count to use for updating modules."
        super().__init__(name=name, default_value=default_value, description=description)

    def verify_value(self, value):
        if not value.isnumeric() or int(value) <= 0:
            raise error.InvalidConfigurationValue(variable=self.name,
                                                  value=value,
                                                  description="Value must be >= 1.")


class BuildJobNumberConfigVariable(BuiltInConfigVariable):
    """Special implementation for the build_job_number configuration variable."""

    def __init__(self):
        name = "build_job_number"
        default_value = "all_but_one"
        description = "Defines the job count to use for building modules.  Value is a number >= " \
                      "1 or any of the following special values: \n" \
                      " - all: Use all cores.\n" \
                      " - all_but_one: Use all cores but one.\n" \
                      " - half: Use half the cores.\n" \
                      " - all_coop: Same as 'all' with cooperative mode.\n" \
                      " - all_but_one_coop: Same as 'all_but_one' with cooperative mode.\n" \
                      " - half_coop: Same as 'half_coop' with cooperative mode.\n" \
                      "Cooperative mode means that a value derived from the one-minute system " \
                      "load will be subtracted compared to the non-cooperative mode."
        super().__init__(name=name, default_value=default_value, description=description)

    def get_allowed_values(self):
        return ["all", "all_but_one", "half", "all_coop", "all_but_one_coop", "half_coop"]

    def verify_value(self, value):
        value = value.lower()

        allowed_values = self.get_allowed_values()

        def invalid_value() -> error.InvalidConfigurationValue:
            description = f"Value must be a number >= 1 or any of: {', '.join(allowed_values)}"
            return error.InvalidConfigurationValue(variable=self.name,
                                                   value=value,
                                                   description=description)

        if value.isnumeric():
            if int(value) <= 0:
                raise invalid_value()
        elif value not in allowed_values:
            raise invalid_value()

    def interpret_value(self, value: str) -> int:
        core_number = multiprocessing.cpu_count()

        self.verify_value(value)

        if value.isnumeric():
            return int(value)

        if "coop" in value:
            load = math.ceil(float(run("cat /proc/loadavg | awk '{ print $1 }'",
                                       quiet=True, capture_output=True)))
            value = value.replace("_coop", "")
        else:
            load = 0

        if value == "all":
            job_number = core_number - load
        elif value == "all_but_one":
            job_number = core_number - 1 - load
        elif value == "half":
            job_number = (core_number // 2) - load
        else:
            job_number = 1

        return max(job_number, 1)


class BuiltInConfigVariables:
    """Built-in config variables in Axii."""

    dev_mode = BuiltInConfigVariable(
        name="dev_mode",
        default_value=False,
        description="Enables development mode."
    )

    print_exception_traceback = BuiltInConfigVariable(
        name="print_exception_traceback",
        default_value=False,
        description="Print exception tracebacks when they occur.  Option only has an effect in "
                    "dev_mode."
    )

    shell_dev_mode = BuiltInConfigVariable(
        name="shell_dev_mode",
        default_value=False,
        description="Enables development mode for the shell bindings."
    )

    axii_gitlab_pat = BuiltInConfigVariable(
        name="axii_gitlab_pat",
        default_value="",
        description="GitLab PAT for the Axii repo for maintainer functionalities."
    )

    default_workspaces_path = BuiltInConfigVariable(
        name="default_workspaces_path",
        default_value="~",
        description="Path to the default location of all workspaces.  By default, this is the home directory."
    )

    auto_activate_workspace = BuiltInConfigVariable(
        name="auto_activate_workspace",
        default_value=False,
        description="Auto-activate a workspace as soon as you change into any of its subdirectories."
    )

    update_by_default = BuiltInConfigVariable(
        name="update_by_default",
        default_value=True,
        description="During upgrade, perform an update on default."
    )

    default_workspace = DefaultWorkspaceConfigVariable()

    git_remote_preferences = GitRemotePreferencesSubconfigVariable()

    update_job_number = UpdateJobNumberConfigVariable()

    build_job_number = BuildJobNumberConfigVariable()

    print_logo = BuiltInConfigVariable(
        name="print_logo",
        default_value=True,
        description="Enables printing the Axii logo and general release information."
    )

    @classmethod
    def get_all_names(cls) -> ty.List[str]:
        """Returns all built-in config variable names."""

        cfg_vars = []
        for name, value in inspect.getmembers(cls):
            if isinstance(value, ConfigVariable):
                cfg_vars.append(name)
        return cfg_vars

    @classmethod
    def get_by_name(cls, name: str) -> BuiltInConfigVariable:
        """Returns a copy of a built-in config variable given its name."""

        return copy.copy(getattr(cls, name))


class ConfigWithBuiltIns(Config):
    """Models a config with Axii built-ins."""

    def __init__(
            self,
            config_vars: ty.Optional[ty.Dict[str, ConfigVariable]] = None
    ):
        super().__init__(config_vars=config_vars)

    @classmethod
    def from_dict(cls, **kwargs) -> "ConfigWithBuiltIns":
        config_vars = {}
        for key, value in kwargs.items():
            cfg_var: ConfigVariable
            if key in BuiltInConfigVariables.get_all_names():
                cfg_var = BuiltInConfigVariables.get_by_name(key)
            else:
                if isinstance(value, dict):
                    cfg_var = SubconfigVariable(name=key)
                else:
                    cfg_var = ConfigVariable(name=key, value_type=type(value))
            cfg_var.set_value(value, parse=True)
            config_vars[key] = cfg_var
        return cls(config_vars=config_vars)

    def get(self, name: str, interpret: bool = True) -> ty.Any:
        if name in self.config_vars:
            return self.config_vars[name].get_value(interpret=interpret)
        if name in BuiltInConfigVariables.get_all_names():
            return BuiltInConfigVariables.get_by_name(name).get_value(interpret=interpret)
        raise error.NoSuchConfigVariable(name)

    def set(self, name: str, value: ty.Any, parse: bool = True):
        if name not in self.config_vars:
            if name in BuiltInConfigVariables.get_all_names():
                cfg_var = BuiltInConfigVariables.get_by_name(name)
            else:
                cfg_var = ConfigVariable(name=name, value_type=type(value))
            self.config_vars[name] = cfg_var
        self.config_vars[name].set_value(value, parse=parse)

    def unset(self, name: str):
        if name in self.config_vars:
            del self.config_vars[name]
        else:
            raise error.NoSuchConfigVariable(name)


class AxiiConfig:
    """Holds Axii's configuration."""

    def __init__(self):
        self.console: Console = StdConsole()
        self.executor = execute

        self.workspace_env_var_name = WORKSPACE_ENV_VAR_NAME
        self.workspace_name_env_var_name = WORKSPACE_NAME_ENV_VAR_NAME

        self.config_dir = os.path.join(xdg_config_home(), "axii")
        os.makedirs(self.config_dir, exist_ok=True)

        self.global_config_file_name = "global_config.json"
        self.global_config_file_path = os.path.join(self.config_dir, self.global_config_file_name)
        self.global_config_backup_file_path = os.path.join(self.config_dir, "global_config_BAK.json")
        self.global_config: ConfigWithBuiltIns = ConfigWithBuiltIns()

        self._known_workspaces_file_name = "known_workspaces.json"
        self._known_workspaces_file_path = os.path.join(self.config_dir,
                                                        self._known_workspaces_file_name)
        self._known_workspaces = {}
        self.load_known_workspaces()

        self._workspace_dir = os.environ.get(WORKSPACE_ENV_VAR_NAME, default=None)
        self._is_workspace_active = self._workspace_dir is not None
        self._workspace_name = None
        if self._is_workspace_active:
            for name, path in self._known_workspaces.items():
                if path == self._workspace_dir:
                    self._workspace_name = name

        self.has_local_config = self._is_workspace_active
        self.local_config_dir: str | None = os.path.join(self._workspace_dir, ".axii") \
            if self.has_local_config else None
        self.local_config_file_name = "config.json"
        self.local_config_file_path: str | None = os.path.join(self.local_config_dir, self.local_config_file_name) \
            if self.has_local_config else None
        self.local_config: Config | None = Config() if self.has_local_config else None

        self._build_job_number = None

        self._create_default_config()

    def is_workspace_active(self) -> bool:
        return self._is_workspace_active

    def get_active_workspace_name(self) -> ty.Optional[str]:
        return self._workspace_name

    def get_known_workspaces(self, reload: bool = False) -> ty.Dict[str, str]:
        if reload:
            self.load_known_workspaces()
        return self._known_workspaces

    def is_axii_locally_installed(self) -> bool:
        import sys
        program_entry_point_path = sys.argv[0]

        if self.is_workspace_active():
            if program_entry_point_path.startswith(self._workspace_dir):
                return True
        return False

    def get_os_codename(self) -> str:
        """Get the OS codename.

        On Debian-based systems, this is equivalent to 'lsb_release -c'.
        """

        def parse_file_to_object(file_path):
            # Create an empty dictionary to hold the key-value pairs
            import re
            data = {}

            # Read the file line by line
            with open(file_path, 'r', encoding="utf-8") as file:
                for line in file:
                    # Use regular expression to extract key-value pairs
                    match = re.match(r'^([^=]+)=(.+)$', line.strip())
                    if match:
                        key = match.group(1)
                        value = match.group(2)
                        # Remove double quotes if present
                        value = value.strip('"')
                        # Add key-value pairs to the dictionary
                        data[key] = value

            return data

        os_codename = "unknown"
        try:
            data = parse_file_to_object("/etc/lsb-release")
            os_codename = data["DISTRIB_CODENAME"] if data["DISTRIB_ID"] == "Ubuntu" else "unknown"
        except:
            pass
        return os_codename

    def load_known_workspaces(self):
        try:
            with open(self._known_workspaces_file_path, encoding="utf-8") as file:
                self._known_workspaces = json.load(file)
        except:
            pass

    def store_known_workspaces(self):
        try:
            with open(self._known_workspaces_file_path, "w", encoding="utf-8") as file:
                json.dump(self._known_workspaces, file, indent=2)
        except:
            pass

    def load_local_config(self):
        """Load variables and values from local config file in workspace."""

        if self.has_local_config:
            os.makedirs(os.path.join(self.local_config_dir), exist_ok=True)
            if not os.path.isfile(self.local_config_file_path):
                return
            with open(self.local_config_file_path, "r", encoding="utf-8") as file:
                self.local_config = Config.from_dict(**json.load(file))

    def load_global_config(self):
        """Load variables and values from global config file."""

        with open(self.global_config_file_path, "r", encoding="utf-8") as file:
            self.global_config = ConfigWithBuiltIns.from_dict(**json.load(file))

    def store_config(self):
        self.store_local_config()
        self.store_global_config()

    def store_local_config(self):
        if self.has_local_config:
            with open(self.local_config_file_path, "w", encoding="utf-8") as file:
                json.dump(self.local_config.to_dict(), file, indent=2)

    def store_global_config(self):
        """Store current variables and values to global config file."""

        with open(self.global_config_file_path, "w", encoding="utf-8") as file:
            json.dump(self.global_config.to_dict(), file, indent=2)

    def create_global_config_backup(self):
        shutil.copyfile(self.global_config_file_path, self.global_config_backup_file_path)

    def restore_global_config_backup(self):
        shutil.copyfile(self.global_config_backup_file_path, self.global_config_file_path)

    def get(self, name, interpret: bool = True, is_global: bool = True):
        """Get the value of a variable."""

        cfg = self._resolve_config(name, is_global=is_global)
        cfg, cfg_name, nested_cfg, nested_name = self._resolve_config_hierarchy(cfg, name)
        if nested_cfg is not None:
            cfg = nested_cfg
            cfg_name = nested_name
        return cfg.get(cfg_name, interpret=interpret)

    def set(self, name: str, value: str, parse: bool = True, is_global: bool = True):
        """Set a variable."""

        cfg = self._resolve_config(name, is_global=is_global, favor_local=True)
        cfg, cfg_name, nested_cfg, nested_name = self._resolve_config_hierarchy(cfg, name)

        if nested_cfg is None:
            cfg.set(cfg_name, value, parse=parse)
        else:
            nested_cfg.set(nested_name, value, parse=parse)
            cfg.set(cfg_name, nested_cfg, parse=False)

    def unset(self, name: str, is_global: bool = True):
        """Unset a variable."""

        cfg = self._resolve_config(name, is_global=is_global)
        cfg, cfg_name, nested_cfg, nested_name = self._resolve_config_hierarchy(cfg, name)
        if nested_cfg is not None:
            cfg = nested_cfg
            cfg_name = nested_name
        cfg.unset(cfg_name)

    def add_known_workspace(self, name, path):
        self._known_workspaces[name] = {"path": path}
        self.store_known_workspaces()

    def remove_known_workspace(self, name):
        del self._known_workspaces[name]
        self.store_known_workspaces()

    def _resolve_config(self, name:str, is_global: bool, favor_local: bool = False):
        if not is_global:
            cfg = None
            if favor_local or self.local_config.has_variable(name):
                return self.local_config
            if cfg is None:
                return self.global_config
        else:
            return self.global_config

    def _resolve_config_hierarchy(self, cfg: Config, name: str) -> ty.Tuple[Config, str, ty.Optional[Config], str]:
        if "." in name:
            cfg_name, nested_name = name.split(".", maxsplit=1)
            return cfg, cfg_name, cfg.get(cfg_name), nested_name
        else:
            return cfg, name, None, ""

    def _create_default_config(self):
        """Ensures that a configuration file with the defaults will be created if not existing."""

        os.makedirs(self.config_dir, exist_ok=True)

        if not os.path.exists(self.global_config_file_path):
            add_to_default = ["default_workspace", "update_job_number", "build_job_number"]
            for name in add_to_default:
                cfg_var = BuiltInConfigVariables.get_by_name(name)
                self.global_config.set(name, cfg_var.default_value)
            self.store_global_config()


axii_config = AxiiConfig()
axii_config.load_local_config()
axii_config.load_global_config()

# Legacy name, some modules still import it via this name
config = axii_config
