import dataclasses as dc
import os
from typing import Dict, List, Optional, Tuple

from armarx_setup.core import error
from armarx_setup.core.module import common
from armarx_setup.core.module.context import Hook
from armarx_setup.core.util import commands
from armarx_setup.core.util.environ_context import EnvironContext


@dc.dataclass
class CMake(Hook):
    project_name: Optional[str] = None
    project_dir: Optional[str] = None

    definitions: Optional[Dict[str, str]] = dc.field(default_factory=dict)
    generator: Optional[Dict] = dc.field(default_factory=dict)
    disable_cmake_warnings: bool = False
    disable_build_warnings: bool = False
    disable_all_warnings: bool = False

    def __post_init__(self):
        if self.project_name is None and self.ctx is not None:
            self.project_name = self.ctx.name.split("/")[-1].replace('-', '_')
        if self.project_dir is not None:
            self.ctx.build_dir = os.path.join(self.ctx.path, self.project_dir, "build")
        if self.disable_all_warnings:
            self.disable_cmake_warnings = True
            self.disable_build_warnings = True

    def generate_requirements(self) -> List[str]:
        return common.CMAKE_DEPENDENCIES

    def generate_env(self, env: "EnvFile"):
        if self.project_name:
            var, val = self.get_cmake_package_dir_var()
            env.add(var, val, by=self.ctx.module_name)

    def __call__(self, *args, verbose=0, **kwargs):
        build_dir = self.ctx.build_dir
        os.makedirs(build_dir, exist_ok=True)

        args = []

        # If conda (python) is used, there might be conflicting libraries.
        # Therefore, force cmake to ignore this directory.
        if "CONDA_PREFIX" in os.environ:
            args.append(os.path.expandvars("-DCMAKE_IGNORE_PATH=$CONDA_PREFIX"))

        # merge local and global cmake generator
        try:
            generator = self.ctx.ws_config.global_.prepare.cmake.generator
        except AttributeError:
            generator = {}
        self.generator = {**generator, **self.generator}

        if self.disable_all_warnings:
            # Disable CMake warnings.
            args += ["-Wno-dev", "-Wno-deprecated"]

        if self.generator and "name" in self.generator:
            args += ["-G", self.generator["name"]]

        # Merge CMake definitions.
        definitions = {
            # These two definitions are required to not have packages export to the global CMake package registry or to
            # at least ignore the global CMake package registry when using `find_package`. See #144 & !657.
            "CMAKE_EXPORT_NO_PACKAGE_REGISTRY": "YES",
            "CMAKE_FIND_USE_PACKAGE_REGISTRY": "NO"
        }
        if self.disable_build_warnings:
            definitions["CMAKE_C_FLAGS"] = "-w"
            definitions["CMAKE_CXX_FLAGS"] = "-w"
        try:
            definitions = {**definitions, **self.ctx.ws_config.global_.prepare.cmake.definitions}
        except AttributeError:
            pass
        definitions = {**definitions, **self.definitions}

        for key, value in definitions.items():
            prefix = "-D"
            if key.startswith(prefix):
                key = key[len(prefix):]
            with EnvironContext(env=self.ctx.env, expand_environment_variables=True):
                value = os.path.expandvars(value)
            args.append(f'{prefix}{key}={value}')

        cmake = ["cmake"] + args + [".."]

        return common.execute_command(command=cmake, cwd=build_dir, env=self.ctx.env)

    def get_cmake_package_dir_var(self) -> Tuple[str, str]:
        return f"{self.project_name}_DIR", f"{self.ctx.build_dir}"

    def test(
            self,
            rerun_failed=False,
            output_on_failure=False,
    ) -> Optional[bool]:
        if os.path.isdir(self.ctx.build_dir):
            try:
                args = ['ctest', "--no-tests=ignore"]
                if rerun_failed:
                    args.append("--rerun-failed")
                if output_on_failure:
                    args.append("--output-on-failure")
                commands.run(" ".join(args), cwd=self.ctx.build_dir)
            except error.CommandFailed:
                error.handle_exception()
                return False
            else:
                return True
