"""Implements the 'extract' strategy for the prepare step."""

import dataclasses as dc
import datetime
import json
import os
import typing as ty

from armarx_setup.core import error
from armarx_setup.core.module import common
from armarx_setup.core.module.context import Hook
from armarx_setup.core.config import axii_config
from armarx_setup.core.util import environ_context
from armarx_setup.core.util.commands import execute as raw_execute


@dc.dataclass
class Extract(Hook):

    file: str
    file_ext: ty.Optional[str] = None
    extract_into: ty.Optional[str] = None

    def __post_init__(self):
        self.extractors = {
            ".tar.gz": self.extract_tar,
            ".tar.xz": self.extract_tar,
            ".tgz": self.extract_tar,
            ".zip": self.extract_zip,
            ".deb": self.extract_deb,
            ".xz": self.extract_xz
        }

        env_ctx = environ_context.EnvironContext(env=self.ctx.env)

        with env_ctx:
            self.file = os.path.expandvars(self.file)

        for ext, _ in self.extractors.items():
            if self.file.endswith(ext):
                self.file_ext = ext
                break
        if not self.file.startswith("/"):
            self.file = os.path.join(self.ctx.path, self.file)
        if self.extract_into is None:
            self.extract_into = self.file[:-len(self.file_ext)]
        else:
            with env_ctx:
                self.extract_into = os.path.expandvars(self.extract_into)

        self.flag_file = os.path.join(self.extract_into, ".axii-data.json")

    def __call__(self, *args, **kwargs) -> common.StepResult:
        return self.extract_file()

    def extract_file(self) -> common.StepResult:
        if not os.path.isfile(self.file):
            axii_config.console.print(f"Archive file '{self.file}' does not exist.")
            return common.StepResult(status=common.StepResultStatus.FAILURE,
                                     summary="Archive does not exist")

        if self._check_flag_file():
            axii_config.console.print(f"Archive file '{self.file}' has already been extracted.")
            return common.StepResult(status=common.StepResultStatus.SKIPPED,
                                     summary="Archive already extracted")

        sub_step_result: common.StepResult = self.extractors[self.file_ext]()

        if sub_step_result.status == common.StepResultStatus.SUCCESS:
            self._write_flag_file()

        if sub_step_result is None:
            sub_step_result = \
                common.StepResult(status=common.StepResultStatus.FAILURE,
                                  summary="Unknown file extension for archive.")

        return sub_step_result

    def extract_tar(self) -> common.StepResult:
        os.makedirs(self.extract_into, exist_ok=True)  # --directory expects an existing directory.

        tar = [
            "tar",
            "--extract",
            "--verbose",
            f"--file={self.file}",  # The .tar.* or .tgz file to extract.
            f"--directory={self.extract_into}"  # The location to extract into.
        ]
        return common.execute_command(
            tar,
            cwd=self.ctx.path,
            success_summary="Successfully extracted archive",
            failure_summary="Failed extracting archive")

    def extract_xz(self):
        os.makedirs(self.extract_into, exist_ok=True)  # unxz is called with cwd set to this dir.

        xzcat = [
            "xzcat",
            "--keep",  # Do not delete .xz file.
            "--no-warn",  # unxz may terminate with exit code 2 in case of warnings.
            self.file,  # The .xz file to extract.
        ]

        stderr: str = ""
        success_summary = "Successfully extracted archive",
        failure_summary = "Failed extracting archive"

        _, file_name = os.path.split(self.file)
        file_name, _ = os.path.splitext(file_name)

        def capture_stderr(text: str):
            nonlocal stderr
            stderr += text + "\n"

        try:
            output = raw_execute(
                xzcat,
                cwd=self.ctx.path,
                stderr_callback=capture_stderr)

            with open(os.path.join(self.extract_into, file_name), "w", encoding="utf-8") as f:
                f.write(output)
        except error.CommandFailed:
            status = common.StepResultStatus.FAILURE
            summary = failure_summary
        except:
            status = common.StepResultStatus.FAILURE
            summary = failure_summary
        else:
            status = common.StepResultStatus.SUCCESS
            summary = success_summary

        return common.StepResult(status=status, summary=summary, executed_command= " ".join(xzcat),
                                 output="", error_output=stderr)

    def extract_zip(self) -> common.StepResult:
        unzip = [
            "unzip",
            self.file,  # The .zip file to extract.
            "-d", self.extract_into  # The location to extract into.
        ]
        return common.execute_command(
            unzip,
            cwd=self.ctx.path,
            success_summary="Successfully extracted archive",
            failure_summary="Failed extracting archive")

    def extract_deb(self) -> common.StepResult:
        dpkg = [
            "dpkg",
            "--vextract",  # Extract verbose (--extract is silent).
            self.file,  # The .deb file to extract.
            self.extract_into  # The location to extract into.
        ]
        return common.execute_command(
            dpkg,
            cwd=self.ctx.path,
            success_summary="Successfully extracted archive",
            failure_summary="Failed extracting archive")

    def _check_flag_file(self) -> bool:
        if not os.path.isfile(self.flag_file):
            return False

        with open(self.flag_file, "r", encoding="utf-8") as file:
            data = json.load(file)

        if self.file in data:
            timestamp = data[self.file]["timestamp"]
        # This is legacy behaviour for workspaces created before 2023-10-16.
        elif "timestamp" in data:
            timestamp = data["timestamp"]

            # Convert file and re-save in new format.
            data[self.file] = {}
            data[self.file]["timestamp"] = timestamp
            del data["timestamp"]
            with open(self.flag_file, "w", encoding="utf-8") as file:
                json.dump(data, file, indent=2)
        else:
            return False

        def fromisoformat(dt_str):
            """Workaround from SO: https://stackoverflow.com/a/28332149 for Python <= 3.7.

            datetime.fromisoformat is only available from Python 3.7 onwards:
            https://docs.python.org/3/library/datetime.html#datetime.datetime.fromisoformat
            """
            dt, _, us = dt_str.partition(".")
            dt = datetime.datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S")
            us = int(us.rstrip("Z"), 10)
            return dt + datetime.timedelta(microseconds=us)

        file_mod_time = datetime.datetime.utcfromtimestamp(os.path.getmtime(self.file))
        # extract_time = datetime.datetime.fromisoformat(data["timestamp"])  # Python >= 3.7
        extract_time = fromisoformat(timestamp)  # Python < 3.7

        # As long as the archive file modification time is smaller than the extract time, we can
        # assume that the extracted files are still valid.
        return file_mod_time < extract_time

    def _write_flag_file(self):
        data = {}
        if os.path.isfile(self.flag_file):
            with open(self.flag_file, "r", encoding="utf-8") as file:
                data = json.load(file)
        with open(self.flag_file, "w", encoding="utf-8") as file:
            timestamp = datetime.datetime.now()
            if self.file not in data:
                data[self.file] = {}
            data[self.file]["timestamp"] = timestamp.isoformat()
            json.dump(data, file, indent=2)
