import dataclasses as dc
import os
import typing as ty

from armarx_setup.core.module import common
from armarx_setup.core.module.context import Hook
from armarx_setup.core.util import environ_context


class Makefile:
    TEMPLATE = """
PACKAGE_DIR ?= {package_dir}
MANIFEST_FILE ?= $(PACKAGE_DIR)/{manifest_file}
MODULE_DEFINITION_FILE ?= {module_definition_file}
MAKEFILE ?= {makefile}

PYTHON_EXEC ?= {python}

VENV_DIR ?= {venv_dir}
VENV_ACTIVATE ?= $(VENV_DIR)/bin/activate
VENV_PYTHON_BIN ?= $(VENV_DIR)/bin/python3
VENV_PIP_BIN ?= $(VENV_DIR)/bin/pip3


all: $(VENV_ACTIVATE)


$(MANIFEST_FILE):


$(MODULE_DEFINITION_FILE):


$(MAKEFILE):


$(VENV_ACTIVATE): $(MANIFEST_FILE) $(MODULE_DEFINITION_FILE) $(MAKEFILE)
    echo "Create virtual environment in $(VENV_DIR) ..."
    "$(PYTHON_EXEC)" -m venv "$(VENV_DIR)"
    echo "Upgrade pip ..."
    "$(VENV_PIP_BIN)" install --upgrade pip
    {preinstall_lines}
    echo "Install package dependencies ..."
    "$(VENV_PIP_BIN)" install -e "$(PACKAGE_DIR)" {install_flags}
    {install_editable_lines}
""".replace("    ", "\t").strip() + "\n"

    PREINSTALL_TEMPLATE = '"$(VENV_PIP_BIN)" install {path}'
    INSTALL_EDITABLE_TEMPLATE = '@"$(VENV_PIP_BIN)" install -e "{path}"'

    @classmethod
    def fill_template(
            cls,
            python: str,
            package_dir: str,
            manifest_file: str,
            module_definition_file: str,
            makefile: str,
            venv_dir: str,
            preinstall: ty.List[str],
            install_editable: ty.List[str],
            install_flags: ty.List[str],
    ) -> str:
        python = os.path.expandvars(python)

        preinstall_lines = cls.make_install_lines(
            packages=preinstall,
            template=cls.PREINSTALL_TEMPLATE,
            message="Preinstall packages ...",
        )
        install_editable_lines = cls.make_install_lines(
            packages=install_editable,
            template=cls.INSTALL_EDITABLE_TEMPLATE,
            message="Install additional packages in editable mode ...",
        )

        install_flags_line = cls.make_install_flags_line(flags=install_flags)

        return cls.TEMPLATE.format(
            python=python,
            package_dir=package_dir,
            manifest_file=manifest_file,
            module_definition_file=module_definition_file,
            makefile=makefile,
            venv_dir=venv_dir,
            preinstall_lines=preinstall_lines,
            install_editable_lines=install_editable_lines,
            install_flags = install_flags_line
        )

    @classmethod
    def make_install_lines(cls, packages: ty.List[str], template: str, message: str) -> str:
        lines = []
        if len(packages) > 0:
            if message:
                lines.append(f'echo "{message}"')
            lines += [
                template.format(path=os.path.expandvars(os.path.expanduser(path)))
                for path in packages
            ]
        return "\n\t".join(lines)

    @classmethod
    def make_install_flags_line(cls, flags: ty.List[str]) -> str:
        return " ".join(["--" + flag.strip().lstrip("--") for flag in flags if flag.strip().lstrip("--") != "-e"])



class SetupPyFile:
    TEMPLATE = """
#!/usr/bin/env python

from distutils.core import setup

setup(name="{name}",
      version="1.0",
      description="{pretty_name}",
      author="Axii Auto-Generator",
      author_email="axii@h2t",
      url="https://git.h2t.iar.kit.edu/",
      packages=[],
      )
"""

    @classmethod
    def fill_template(cls, name) -> str:
        words = name.split("_")
        words = [word.capitalize() for word in words]
        pretty_name = " ".join(words)
        return cls.TEMPLATE.format(name=name, pretty_name=pretty_name)


@dc.dataclass
class PythonPackage(Hook):
    relative_path: str

    env: ty.Dict[str, str] = dc.field(default_factory=dict)
    python: str = "/usr/bin/python3"
    venv_name: str = ".venv"
    manifest_file: str = None

    install_flags: ty.List[str] = dc.field(default_factory=list)
    preinstall: ty.List[str] = dc.field(default_factory=list)
    install_editable: ty.List[str] = dc.field(default_factory=list)
    auto_generate_setup_py: bool = False

    def __call__(self, *args, verbose=0, ignore_errors=False, **kwargs) -> common.StepResult:
        """
        We create a Makefile to create the python virtual environment.
        It is set up to only run pip if the environment needs to be updated.
        """
        from armarx_setup import console

        console.print(f"\nPrepare Python package '{self.relative_path}' in module "
                      f"'{self.ctx.name_colored}' ...")

        python_package_path = os.path.join(self.ctx.path, self.relative_path)
        venv_dir = os.path.join(python_package_path, self.venv_name)
        manifest_file = self.manifest_file or self.find_manifest_file(python_package_path)

        # Create the venv directory, if it does not exist.
        os.makedirs(venv_dir, exist_ok=True)

        if self.auto_generate_setup_py:
            self.generate_and_write_setup_py()

        # Create the Makefile if it does not exist.
        # We put it _into_ the venv, so it is ignored automatically.
        makefile_filename = "Makefile"
        makefile_path = os.path.join(venv_dir, makefile_filename)

        env = {**self.ctx.env, **self.env}

        with environ_context.EnvironContext(env=env, expand_environment_variables=True):
            makefile_content = Makefile.fill_template(
                    python=self.python,
                    package_dir="..",
                    manifest_file=manifest_file,
                    module_definition_file=self.ctx.module_info_path,
                    makefile=makefile_filename,
                    venv_dir=".",
                    preinstall=self.preinstall,
                    install_editable=self.install_editable,
                    install_flags=self.install_flags
                )

        # Makefile recreation rules.

        def file_not_created():
            return not os.path.exists(makefile_path)

        def code_newer_than_makefile():
            # Did this code change after the makefile was created?
            mod_time_makefile = os.path.getmtime(makefile_path)
            mod_time_code = os.path.getmtime(os.path.abspath(__file__))
            return mod_time_makefile < mod_time_code

        def content_changed():
            assert os.path.isfile(makefile_path), makefile_path
            with open(makefile_path, "r", encoding="utf-8") as file:
                current_content = file.read()
            return current_content != makefile_content

        recreate_rules = [
            file_not_created,
            code_newer_than_makefile,
            content_changed,
        ]
        recreate_makefile = any(rule() for rule in recreate_rules)

        if recreate_makefile:
            console.print(f"Generate Makefile in '{venv_dir}' ...")
            # Dirs are relative to the venv.
            with open(makefile_path, "w", encoding="utf-8") as file:
                file.write(makefile_content)

        # Call make.
        return common.execute_command(["make"], cwd=os.path.dirname(makefile_path), env=env)

    def generate_and_write_setup_py(self):
        setup_py_path = self.ctx.path
        setup_py_file = os.path.join(setup_py_path, "setup.py")
        generated_package_name = self.ctx.name.replace("/", "_")
        setup_py_contents = SetupPyFile.fill_template(generated_package_name)

        with open(setup_py_file, "w", encoding="utf-8") as file:
            file.write(setup_py_contents)

    def test(self, **kwargs):
        pass

    @classmethod
    def find_manifest_file(cls, python_package_path: str) -> ty.Optional[str]:
        for manifest_filename in ["pyproject.toml", "requirements.txt"]:
            if os.path.isfile(os.path.join(python_package_path, manifest_filename)):
                return manifest_filename
        return None


@dc.dataclass
class Python(Hook):
    packages: ty.Dict[str, PythonPackage] = dc.field(default_factory=dict)

    def __post_init__(self):
        from armarx_setup.core.util.dataclass import sanitize_type
        self.packages = {
            path: sanitize_type(dict(relative_path=path, **data), PythonPackage, ctx=self.ctx)
            for path, data in self.packages.items()}

    def generate_requirements(self) -> ty.List[str]:
        return [
            "apt/python3-venv",
        ]

    def clean_venvs(self):
        import shutil
        for package in self.packages.values():
            venv_path = os.path.normpath(os.path.join(self.ctx.path, package.relative_path, package.venv_name))
            if os.path.exists(venv_path):
                shutil.rmtree(venv_path)

    def __call__(self, *args, **kwargs) -> common.StepResult:
        step_results = []
        for package in self.packages.values():
            step_result = package(**kwargs)
            step_results.append(step_result)
        return common.StepResult.from_sub_steps_results(sub_steps_results=step_results)

    def test(self, **kwargs):
        pass
