import os

import dataclasses as dc
import typing as ty

from armarx_setup.core.module.context import Hook
from armarx_setup.core.util import commands, environ_context
from armarx_setup.core.module import common


@dc.dataclass
class Download(Hook):

    url: str
    file: ty.Optional[str] = None
    md5_checksum: ty.Optional[str] = None
    make_executable: bool = False

    def __post_init__(self):
        env_ctx = environ_context.EnvironContext(env=self.ctx.env)

        with env_ctx:
            self.url = os.path.expandvars(self.url)

        if self.file is None:
            _, self.file = os.path.split(self.url)
        with env_ctx:
            self.file = os.path.expandvars(self.file)
        if self.file.startswith("/"):
            self.full_path = self.file
        else:
            self.full_path = os.path.join(self.ctx.path, self.file)

    def __call__(self, *args, **kwargs):
        return self.download_file()

    def download_file(self):
        if os.path.exists(self.full_path):
            return common.StepResult(status=common.StepResultStatus.SKIPPED,
                                     summary="File already downloaded",
                                     output=f"File '{self.full_path}' has already been downloaded.")

        tmp_file = f"{self.file}~"

        # Ensure that the folder to download into exists.
        download_path, _ = os.path.split(self.full_path)
        os.makedirs(download_path, exist_ok=True)

        tmp_full_path = f"{self.full_path}~"

        # Remove any remains of potential previous failed runs.
        if os.path.exists(tmp_full_path):
            os.remove(tmp_full_path)

        wget = [
            "wget",
            "--tries=2",  # 2 retries for increased robustness.
            f"--output-document={tmp_file}",
            self.url
        ]

        sub_step_result = common.execute_command(
            wget,
            cwd=self.ctx.path,
            success_summary="Downloaded file",
            failure_summary="Failed downloading file",
            executor=commands.execute  # Use plain executor to fully capture output.
        )

        def filter_wget_stderr(stderr: str) -> str:
            lines = stderr.strip().split("\n")
            new_lines = []
            for line in lines:
                if any(f"{i}%[" in line for i in range(101)):
                    continue
                new_lines.append(line.replace("\r", ""))
            return "\n".join(new_lines)

        # wget only outputs to stderr.
        sub_step_result.output = filter_wget_stderr(sub_step_result.error_output)

        if sub_step_result.status == common.StepResultStatus.FAILURE:
            return sub_step_result
        else:
            # Perform MD5 checksum verification if defined.
            if self.md5_checksum is not None:
                actual_md5_checksum = self.get_actual_md5_checksum(tmp_file)

                report = f"MD5 checksum of '{tmp_file}': {actual_md5_checksum}\n" \
                         f"MD5 checksum expected: {self.md5_checksum}\n"
                sub_step_result.output += f"\n{report}"
                sub_step_result.error_output += f"\n{report}"

                if actual_md5_checksum != self.md5_checksum:
                    sub_step_result.status = common.StepResultStatus.FAILURE
                    sub_step_result.summary = "MD5 checksum verification failed"
                    return sub_step_result

            if self.make_executable and not os.access(tmp_full_path, os.X_OK):
                os.chmod(tmp_full_path, 0o750)

            os.rename(tmp_full_path, self.full_path)

        return sub_step_result

    def get_actual_md5_checksum(self, tmp_file: str):
        step_result = common.execute_command(
            ["md5sum", tmp_file],
            cwd=self.ctx.path,
            executor=commands.execute)

        return step_result.output.split(" ")[0]
