"""Implements the Git strategy for the update step."""

import abc
import dataclasses as dc
import os
import typing as ty

from armarx_setup.core import error
from armarx_setup.core.module import common
from armarx_setup.core.module.context import Hook
from armarx_setup.core.util import commands
from armarx_setup.core.config import axii_config, GitRemotePreferencesSubconfig


git_non_interaction_env_vars = {
    "GIT_SSH_COMMAND": "ssh -oBatchMode=yes",  # Disables interaction in ssh.
    "GIT_TERMINAL_PROMPT": "0",  # Disables interaction in Git, such as HTTPS credentials.
}


class GitCodeHoster:

    @abc.abstractmethod
    def domain(self) -> str:
        ...

    @abc.abstractmethod
    def home_url(self):
        ...

    @abc.abstractmethod
    def ssh_clone_url(self):
        ...

    @abc.abstractmethod
    def https_clone_url(self):
        ...

    @abc.abstractmethod
    def issues_url(self):
        ...

    @abc.abstractmethod
    def merge_requests_url(self):
        ...


@dc.dataclass
class GitLab(GitCodeHoster):
    slug: str
    instance_url: str

    def domain(self) -> str:
        return self.instance_url

    def home_url(self):
        return f"https://{self.domain()}/{self.slug}"

    def ssh_clone_url(self):
        return f"git@{self.domain()}:{self.slug}.git"

    def https_clone_url(self):
        return f"https://{self.domain()}/{self.slug}.git"

    def issues_url(self):
        return os.path.join(self.home_url(), "-", "issues")

    def new_issue_url(self):
        return os.path.join(self.issues_url(), "new")

    def merge_requests_url(self):
        return os.path.join(self.home_url(), "-", "merge_requests")

    def releases_url(self, release = ""):
        return os.path.join(self.home_url(), "-", "releases", release)


@dc.dataclass
class GitLabCom(GitLab):
    instance_url: str = "gitlab.com"


@dc.dataclass
class H2TGitLab(GitLab):
    instance_url: str = "git.h2t.iar.kit.edu"


@dc.dataclass
class GitHub(GitCodeHoster):
    slug: str

    def domain(self) -> str:
        return "github.com"

    def home_url(self):
        return f"https://github.com/{self.slug}"

    def ssh_clone_url(self):
        return f"git@github.com:{self.slug}.git"

    def https_clone_url(self):
        return f"https://github.com/{self.slug}.git"

    def issues_url(self):
        return os.path.join(self.home_url(), "issues")

    def merge_requests_url(self):
        return os.path.join(self.home_url(), "pulls")


@dc.dataclass
class Git(Hook):

    ssh_url: ty.Optional[str] = None
    https_url: ty.Optional[str] = None

    gitlab_instance: ty.Optional[str] = None
    gitlab_slug: ty.Optional[str] = None
    gitlab_com_slug: ty.Optional[str] = None
    h2t_gitlab_slug: ty.Optional[str] = None
    github_slug: ty.Optional[str] = None

    default_checkout_ref: ty.Optional[str] = None
    default_checkout_commit: ty.Optional[str] = None
    submodules: bool = False
    lfs: ty.Optional[bool] = None
    ignore_patterns: ty.Optional[ty.List[str]] = None
    prefer_https: bool = False

    def __post_init__(self):
        git_related_fields = {self.gitlab_instance, self.gitlab_slug, self.gitlab_com_slug,
                              self.h2t_gitlab_slug, self.github_slug, self.https_url, self.ssh_url}

        non_generic_gitlab = git_related_fields - {self.gitlab_instance, self.gitlab_slug}
        non_gitlab_com = git_related_fields - {self.gitlab_com_slug}
        non_h2t_gitlab = git_related_fields - {self.h2t_gitlab_slug}
        non_github = git_related_fields - {self.github_slug}

        self.code_hoster: ty.Optional[GitCodeHoster] = None
        if self.gitlab_instance is not None and self.gitlab_slug is not None:
            assert all(x is None for x in non_generic_gitlab)
            self.code_hoster = GitLab(instance_url=self.gitlab_instance, slug=self.gitlab_slug)
        elif self.gitlab_com_slug is not None:
            assert all(x is None for x in non_gitlab_com)
            self.code_hoster = GitLabCom(slug=self.gitlab_com_slug)
        elif self.h2t_gitlab_slug is not None:
            assert all(x is None for x in non_h2t_gitlab)
            self.code_hoster = H2TGitLab(slug=self.h2t_gitlab_slug)
        elif self.github_slug is not None:
            assert all(x is None for x in non_github)
            self.code_hoster = GitHub(slug=self.github_slug)

        if self.code_hoster is not None:
            self.https_url = self.code_hoster.https_clone_url()
            self.ssh_url = self.code_hoster.ssh_clone_url()

    def __call__(self, prefer_https: bool = False, **kwargs) -> common.StepResult:
        if os.path.isdir(os.path.join(self.ctx.path, ".git")):
            result = self.pull(**kwargs)
        else:
            result = self.clone(prefer_https=self.prefer_https or prefer_https, **kwargs)

        self.update_ignore_patterns()
        return result

    def clone(self, prefer_https: bool = False, **kwargs) -> common.StepResult:
        """Clones a Git repository."""

        url = self.https_url

        if not prefer_https and self.code_hoster is not None:
            grp: GitRemotePreferencesSubconfig = axii_config.get("git_remote_preferences")
            if grp.preference_for_host(self.code_hoster.domain()) == "ssh":
                url = self.ssh_url

        args = [url, self.ctx.path]
        if self.default_checkout_ref:
            args += ["--branch", self.default_checkout_ref]
        if self.submodules:
            args.append("--recursive")

        git_clone = ["git", "clone", "--progress", "--verbose"] + args

        sub_step_result = common.execute_command(
            git_clone,
            cwd=self.ctx.ws_path,
            env=git_non_interaction_env_vars,
            success_summary="Cloned repository",
            failure_summary="Failed cloning repository",
            executor=commands.execute  # Use plain executor to fully capture output.
        )

        if self.default_checkout_commit:
            git_checkout = ["git", "checkout", self.default_checkout_commit]
            checkout_sub_step_result = common.execute_command(
                git_checkout,
                cwd=self.ctx.path,
                env=git_non_interaction_env_vars,
                executor=commands.execute  # Use plain executor to fully capture output.
            )

            if checkout_sub_step_result.status == common.StepResultStatus.FAILURE:
                return checkout_sub_step_result

        def filter_git_clone_stderr(stderr: str) -> str:
            lines = stderr.strip().split("\n")
            new_lines = []
            blocked_phrases = [
                "Enumerating objects:",
                "Counting objects:",
                "Compressing objects:",
                "Receiving objects:",
                "Resolving deltas:",
                "Updating files:",
                "Checking out files:",
                "Filtering content:",
                "Finding sources:"]
            for line in lines:
                if any(blocked in line for blocked in blocked_phrases) and ", done." not in line:
                    continue
                new_lines.append(line.replace("\r", ""))
            return "\n".join(new_lines)

        # Git prints all output to stderr in `git clone`.  Here we filter all verbose updates
        # and forward the essentials to stdout.
        sub_step_result.output = filter_git_clone_stderr(sub_step_result.error_output)

        return sub_step_result

    def pull(self, **kwargs) -> common.StepResult:
        """Pulls a Git repository."""

        import git

        detached_head_state: ty.Optional[str] = self.get_detached_head_state()

        if detached_head_state is None:
            try:
                repo = git.Repo(self.ctx.path)
            except git.GitError as e:
                return common.StepResult(
                    status=common.StepResultStatus.FAILURE,
                    summary="Exception initializing Git repository object",
                    error_output=f"Exception '{repr(e)}' occurred while initializing a Git "
                                 f"repository object")
            active_branch_name: str = repo.active_branch.name
            active_branch = f"'[cyan b]{active_branch_name}[/]'"

            try:
                git_pull = "git pull --prune --no-edit".split()
                stdout = commands.execute(git_pull, cwd=self.ctx.path,
                                          env=git_non_interaction_env_vars)

                if self.submodules:
                    git_submodules_update = "git submodule update --recursive".split()
                    sm_stdout = commands.execute(git_submodules_update, cwd=self.ctx.path,
                                                 env=git_non_interaction_env_vars)
                    if sm_stdout != "":
                        stdout += f"\n{sm_stdout}"

            except error.CommandFailed as e:
                stdout = e.stdout
                stderr = e.stderr

                stdout = stdout + "\n" + stderr

                status = common.StepResultStatus.FAILURE

                if "You are not currently on a branch" in stdout:
                    summary = ("update.Git.get_detached_head_state() was None but still "
                               "'git pull' resulted in 'not on a branch' state. This should not "
                               "have happened.")
                elif "There is no tracking information for the current branch" in stdout:
                    status = common.StepResultStatus.SUCCESS_WITH_WARNING
                    summary = "Cannot pull from a local branch"
                elif "Your local changes to the following files would be overwritten by merge" \
                        in stdout:
                    summary = f"Local changes would be overwritten (on branch {active_branch})."
                else:
                    summary = "Update step failed"

                return common.StepResult(status=status, summary=summary,
                                         executed_command="git pull && git submodules update",
                                         output=stdout, error_output=stderr)
            else:
                if "Already up-to-date" in stdout or "Already up to date" in stdout:
                    summary = f"Up to date on branch {active_branch}"
                elif "Updating" in stdout:
                    summary = f"Updated repository on branch {active_branch}"
                else:
                    summary = "Update step executed successfully"

                return common.StepResult(status=common.StepResultStatus.SUCCESS, summary=summary,
                                         executed_command="git pull && git submodules update",
                                         output=stdout)
        else:
            return common.StepResult(
                status=common.StepResultStatus.SKIPPED,
                summary="Nothing to do, on tag or commit",
                output=f"Skipped update because repository is on {detached_head_state}."
            )

    def get_detached_head_state(self) -> ty.Optional[str]:
        kwargs = {
            "quiet": True,
            "cwd": self.ctx.path,
            "capture_output": True
        }

        try:
            head_hash = commands.run("git rev-parse HEAD", **kwargs)
            head_hash_short = commands.run("git rev-parse --short HEAD", **kwargs)

            # Check if repo is on a user-specified tag (no error).
            try:
                tag = commands.run("git name-rev --name-only --tags HEAD", **kwargs)
                if tag == self.default_checkout_ref:
                    current_ref_hash = commands.run(
                        f"git rev-list -n 1 {self.default_checkout_ref}", **kwargs)
                    if current_ref_hash == head_hash:
                        return f"tag '{self.default_checkout_ref}'"
            except error.CommandFailed:
                pass

            # Check if repo is on a user-specified commit.
            res = commands.run("git rev-parse --abbrev-ref HEAD", **kwargs)
            if res == "HEAD":
                return f"commit '{head_hash_short}'"

        except error.CommandFailed:  # as e:
            # print("command failed", e)
            pass

    def update_ignore_patterns(self):
        git_folder_name = ".git"
        ignore_filename = f"{git_folder_name}/info/exclude"
        git_folder_filepath = os.path.join(self.ctx.path, git_folder_name)
        ignore_filepath = os.path.join(self.ctx.path, ignore_filename)

        # If the .git folder does not exist yet, do nothing (if git clone fails etc.).
        if not os.path.isdir(git_folder_filepath):
            return

        # If the exclude file does not exist yet, create an empty one.
        if not os.path.isfile(ignore_filepath):
            info_path, _ = os.path.split(ignore_filepath)
            os.makedirs(info_path, exist_ok=True)
            with open(ignore_filepath, "a", encoding="utf-8"):
                pass

        # Read current file.
        with open(ignore_filepath, "r", encoding="utf-8") as file:
            lines = file.readlines()
        lines: ty.List[str] = [line.rstrip("\n") for line in lines]

        # Apply required changes.
        changed = False
        if self.ignore_patterns is not None:
            for pattern in self.ignore_patterns:
                if pattern not in lines:
                    lines.append(pattern)
                    changed = True

        # Write new file if necessary.
        if changed:
            lines = [line + "\n" for line in lines]
            with open(ignore_filepath, "w", encoding="utf-8") as file:
                file.writelines(lines)
