"""Manages a workspace."""

import os
import re
import typing as ty
from typing import Any, Callable, Dict, List, Optional

from rich.rule import Rule

from armarx_setup import console, PACKAGE_ROOT
from armarx_setup.core import error
from armarx_setup.core.config import axii_config
from armarx_setup.core.module.common import StepResult, StepResultStatus
from armarx_setup.core.workspace.home import Home


AXII_WORKSPACE_OS_FINGERPRINT = "AXII_WORKSPACE_OS_FINGERPRINT"


class Workspace:
    """An object of this class allows interacting with and managing a workspace.
    """

    def __init__(
            self,
            name: str,
            path: str,
            config: Optional["Config"] = None,
            enable_env_file=True,
    ):
        from armarx_setup.core.workspace.config import Config
        from armarx_setup.core.workspace.env import EnvFile

        if path.endswith("/"):
            path = path[:-len("/")]

        self.name = name
        self.path = path
        self.config = config or Config()
        self._home = Home(workspace_path=path)

        if enable_env_file:
            self.env_file = EnvFile(ws_path=self.path)
            self.env_file.add(axii_config.workspace_env_var_name, path, by="workspace", quiet=True)
            self.env_file.add(axii_config.workspace_name_env_var_name, self.name, by="workspace", quiet=True)
            self.env_file.add(AXII_WORKSPACE_OS_FINGERPRINT, axii_config.get_os_codename(), by="workspace", quiet=True)
            self.env_file.add("ARMARX_USER_CONFIG_DIR", os.path.join(path, "armarx_config"), by="workspace", quiet=True)
        else:
            self.env_file = None

        self._module_graph: Optional["ModuleGraph"] = None

    @classmethod
    def load_workspace_by_name(cls, name: str):
        known_workspaces = axii_config.get_known_workspaces()
        try:
            path = known_workspaces[name]["path"]
        except KeyError as e:
            raise error.NoSuchWorkspace(workspace_name=name) from e
        else:
            ws = Workspace(name=name, path=path)
            ws.load_config()
            return ws

    @classmethod
    def load_workspace_by_path(
            cls,
            path: str
    ) -> "Workspace":
        if not os.path.isdir(path):
            raise error.NoSuchWorkspace(workspace_name=path)

        name = path
        try:
            known_workspaces = axii_config.get_known_workspaces()
            for known_workspace_name, known_workspace in known_workspaces.items():
                if known_workspace["path"] == path:
                    name = known_workspace_name
                    break
            else:  # If workspace was not found in known_workspaces.json, add it with name=path.
                axii_config.add_known_workspace(name=path, path=path)
        except Exception as e:
            print(e)

        ws = Workspace(name=name, path=path)
        ws.load_config()
        return ws

    @classmethod
    def get_active_workspace_dir(cls):
        path = os.environ.get(axii_config.workspace_env_var_name, None)
        if not path:
            raise error.NoWorkspaceActive()
        path = os.path.expandvars(os.path.expanduser(path))
        return path

    @classmethod
    def is_workspace_active(cls):
        return os.environ.get(axii_config.workspace_env_var_name, None) is not None

    @classmethod
    def load_active_workspace(cls, enforce_os_fingerprint=True):
        path = cls.get_active_workspace_dir()

        workspace_os_fingerprint = os.environ.get(AXII_WORKSPACE_OS_FINGERPRINT, None)
        actual_os_fingerprint = axii_config.get_os_codename()
        if (enforce_os_fingerprint and
                workspace_os_fingerprint is not None and
                workspace_os_fingerprint != actual_os_fingerprint):
            raise error.WorkspaceWasBuiltWithOtherOS(workspace_os_fingerprint=workspace_os_fingerprint,
                                                     actual_os_fingerprint=actual_os_fingerprint)

        return cls.load_workspace_by_path(path)

    @staticmethod
    def get_all_workspace_names():
        return axii_config.get_known_workspaces().keys()

    def create(self):
        """Create directories and store the config."""

        import collections
        info = collections.OrderedDict()

        name = self.name
        path = self.path
        config_path = self.get_config_path(path)

        info["Workspace name"] = name
        info["Workspace path"] = path

        os.makedirs(path, exist_ok=True)
        assert os.path.isdir(path), path

        if os.path.isfile(self.get_config_path()):
            raise error.WorkspaceAlreadyExists(path)
        if len(os.listdir(path)) > 0:
            raise error.AxiiError(f"Path {path} is not empty")

        info["Workspace config"] = config_path

        self.store_config()

        info["Workspace environment file"] = self.env_file.get_path()

        self.env_file.write()

        axii_config.add_known_workspace(name=name, path=path)

        return info

    def load_config(self) -> "Config":
        from armarx_setup.core.workspace.config import Config

        try:
            self.config = Config.load(self.get_config_path())
        except IOError:
            raise
        return self.config

    def store_config(self):
        self.config.store(self.get_config_path())

    def get_config_path(self, path=None):
        if path is None:
            path = self.path
        return os.path.join(path, "armarx-workspace.json")

    def get_local_axii_config_path(self, path=None):
        if path is None:
            path = self.path
        return os.path.join(path, ".axii")

    def add_module(
            self,
            module_name: str,
            check=True,
            store=True,
    ):
        from armarx_setup.core.module import Module

        module_name = module_name.strip("/")
        if module_name in self.config.modules:
            console.print(f"Module '{module_name}' already added.")
            return False

        if check:
            Module.load_info(module_name, ws=self)

        self.config.modules[module_name] = {}
        self._module_graph = None
        if store:
            self.store_config()
            console.print(f"Added module '{module_name}'.")

        return True

    def add_module_features(
            self,
            module_name: str,
            feature_names: List[str],
            check=True,
            store=True,
    ):
        from armarx_setup.core.module import Module

        if module_name not in self.config.modules:
            if self.add_module(module_name=module_name, store=False):
                console.print(f"Added module '{module_name}'.")
            else:
                return  # Failed to add module.

        if check:
            module: Module = Module.load_info(module_name)
            for feature_name in feature_names:
                if feature_name not in module.features:
                    console.print(f"Module '{module_name}' does not provide a feature named "
                                  f"'{feature_name}'.")

                    similar = Module.get_similar_module_names(feature_name, module.features.keys())
                    if similar:
                        console.print("\tSimilar feature names: " +
                                      ", ".join(f"'{c}'" for c in similar))

                    return False

        module_dict: Dict = self.config.modules[module_name]
        feature_names = set(feature_names)
        current_features = set(module_dict.get("features", []))

        merged_features = current_features.union(feature_names)
        already_added = current_features.intersection(feature_names)
        new_features = feature_names.difference(current_features)

        def str_feature_numerus(features):
            subject = "Feature"
            are = "is"
            if len(features) > 1:
                subject += "s"
                are = "are"
            return subject, are

        def str_feature_names(features) -> str:
            return ", ".join(f"'{f}'" for f in features)

        def str_already_added(features):
            subject, are = str_feature_numerus(features)
            return f"{subject} {str_feature_names(features)} of module '{module_name}' {are} " \
                   f"already added."

        if len(new_features) == 0:
            console.print(str_already_added(feature_names))
            return False

        module_dict["features"] = list(merged_features)
        if store:
            self.store_config()
            subject, _ = str_feature_numerus(new_features)
            if already_added:
                console.print(str_already_added(already_added))
            console.print(f"Added {subject.lower()} {str_feature_names(new_features)} of module "
                          f"'{module_name}'.")

        return True

    def remove_module_features(self, module_name: str, feature_names: List[str], store=True):
        if module_name not in self.config.modules:
            console.print(f"Module '{module_name}' is not in the current workspace.")
            return

        module_dict: Dict = self.config.modules[module_name]
        current_features = set(module_dict.get("features", []))

        def str_feature_numerus(features):
            subject = "Feature"
            are = "is"
            if len(features) > 1:
                subject += "s"
                are = "are"
            return subject, are

        removing_features = set(feature_names)
        invalid_features = removing_features.difference(current_features)
        if invalid_features:
            subject, are = str_feature_numerus(invalid_features)
            console.print(f"{subject} {', '.join(invalid_features)} {are} not active in the workspace.")
            return

        module_dict["features"] = list(current_features.difference(feature_names))
        if not module_dict["features"]:
            del module_dict["features"]

        self._module_graph = None

        if store:
            self.store_config()
            subject, _ = str_feature_numerus(removing_features)
            console.print(f"Removed {subject.lower()} {', '.join(removing_features)} of module "
                          f"'{module_name}'.")
        return True

    def remove_module(self, module_name: str, store=True):
        module_name = module_name.strip("/")
        if module_name not in self.config.modules:
            console.print(f"Module '{module_name}' is not in the current workspace.")

            from armarx_setup.core.module import Module
            similar = Module.get_similar_module_names(module_name, self.config.modules)
            if similar:
                console.print("Similar module names: " + ", ".join(f"'{c}'" for c in similar))

        else:
            del self.config.modules[module_name]
            self._module_graph = None
            if store:
                self.store_config()
                console.print(f"Removed module '{module_name}'.")

    @property
    def module_graph(self) -> "ModuleGraph":
        if self._module_graph is None:
            self._module_graph = self.build_module_graph()
        return self._module_graph

    @module_graph.setter
    def module_graph(self, value: "ModuleGraph"):
        self._module_graph = value

    def build_module_graph(self) -> "ModuleGraph":
        from armarx_setup.core.module_graph import ModuleGraph

        # console.print("Build module graph ...", end="")
        graph = ModuleGraph.from_modules(self.config.modules, ws=self)
        # console.print(" Done.")
        return graph

    def for_each_module_in_dependency_chain(
            self, func: Callable[..., Any], *args, **kwargs
    ) -> Dict[str, Any]:
        return {module.name: func(module, *args, **kwargs)
                for module in self.module_graph.get_dependency_chain()}

    def get_modules_in_dependency_chain(self) -> List["Module"]:
        return list(self.module_graph.get_dependency_chain())

    def for_each_module_in_alphabetical_order(
            self, func: Callable[..., Any], *args, **kwargs
    ) -> Dict[str, Any]:

        modules = list(self.module_graph.nodes(data="module"))
        modules.sort()

        return {module.name: func(module, *args, **kwargs)
                for _, module in modules}

    def get_modules_in_alphabetical_order(self) -> List["Module"]:
        return [module for _, module in self.module_graph.nodes(data="module")]

    def info(self, graph=False, layout=None, reduce=False, all=False):
        from armarx_setup.core.util import prefix_color_map

        def decorate_module(module: "Module") -> str:
            if isinstance(module, str):
                module = self.module_graph.nodes[module]["module"]
            module_name = prefix_color_map.module_to_colored_text(module.name)
            if module.general is not None and module.general.deprecated is not None:
                module_name = f"[strike]{module_name}[/strike]"
            elif module.not_found:
                module_name = f"[bold red blink]{module.name}[/] (module definition not found)"
            return module_name

        def make_table():
            from rich.table import Table
            table = Table("#", "Module", "Required Modules",
                          title="Added Modules and their Direct Dependencies")
            table.columns[0].justify = "right"
            for i, module_name in enumerate(sorted(self.config.modules)):
                required = sorted(self.module_graph.nodes[module_name]["module"].required_modules)
                if not all:
                    required = self.module_graph.filter_hidden_modules(required)
                required_modules = ", ".join(decorate_module(m) for m in required)
                if not required_modules:
                    required_modules = "[italic]none[/italic]"
                table.add_row(str(i), decorate_module(module_name), required_modules)
            return table

        table = make_table()
        console.print(table)
        console.print()

        def make_table():
            from rich.table import Table
            table = Table("#", "Module", "Required by", title="Complete List of Modules")
            table.columns[0].justify = "right"
            for i, node in enumerate(sorted(self.module_graph.nodes)):
                module = self.module_graph.nodes[node]["module"]

                if not all and self.module_graph.is_hidden_module(module.name):
                    continue

                required_by = {self.module_graph.nodes[f]["module"].name: d for f, _, d in
                               self.module_graph.in_edges(node, data="dependency")}

                required_by_list = []
                if module.name in self.config.modules:
                    required_by_list = ['[bold]Workspace config[/bold]']

                for dependency_name, dependency in required_by.items():
                    text = decorate_module(dependency_name)
                    if dependency.features:
                        features = ', '.join(f'{f}' for f in dependency.features)
                        text += f' (with features: [italic]{features}[/italic])'
                    elif dependency.feature_name is not None:
                        text += f' (from feature [italic]{dependency.feature_name}[/italic])'
                    required_by_list.append(text)

                table.add_row(str(i), decorate_module(module), ", ".join(required_by_list))
            return table

        table = make_table()
        console.print(table)

        import networkx as nx
        try:
            chain = self.module_graph.get_dependency_chain()
        except nx.NetworkXUnfeasible as e:
            console.print(f"Could not infer dependency chain due to cyclic dependencies: {e}")
        else:
            if not all:
                chain = self.module_graph.filter_hidden_modules(chain)

            text_chain = " :right_arrow: ".join(decorate_module(c) for c in chain)
            console.print()
            console.print(f"Flattened dependency chain:\n{text_chain}")

        if not all:
            console.print()
            console.print("This is a condensed view excluding some types of modules (e.g. pure "
                          "system modules).  To include them, re-run with '--all'.")

        if graph:
            console.print()
            console.print("Show module graph ...")
            self.module_graph.visualize(all=all, layout=layout, reduce=reduce)

    def get_modules_with_system(self, **kwargs) -> ty.List["Module"]:
        def enlist(module: "Module"):
            if module.system is not None:
                return module

        verb = "setup the system for"
        modules = [m for m in self._step_with_options(enlist, verb=verb, **kwargs).values()
                   if m is not None]

        return modules

    def system(
            self,
            accept_apt_install=False,
            **kwargs,
    ) -> ty.Dict[str, ty.Optional[StepResult]]:
        from armarx_setup.core.util.prefix_color_map import modules_to_colored_text as colorize
        from armarx_setup.core.workspace.system import System, SystemStep

        # First gather all apt packages required by modules, then perform the install step.
        modules = self.get_modules_with_system(**kwargs)

        system = System(ws=self)
        steps = system.plan_steps(modules)

        def format_modules(step):
            modules_col = colorize([m.name for m in step.modules])
            return ", ".join(modules_col)

        verbose = 1
        if verbose >= 1:
            console.print("[b]System steps:[/]")
            for i, step in enumerate(steps):
                console.print(f"[{i + 1}] {step.type.name + ':':<9}", format_modules(step))

        step_results = {}

        for step in steps:

            if step.type == SystemStep.Type.Run:
                console.print("Run system scripts for modules", format_modules(step))
                step.run()
                for module in step.modules:
                    step_results[module.name] = \
                        StepResult(status=StepResultStatus.SUCCESS,
                                   summary="Ran system scripts for all modules")

            elif step.type == SystemStep.Type.Packages:
                console.print("Install system packages for modules", format_modules(step))
                step.install_packages(accept=accept_apt_install)
                for module in step.modules:
                    step_results[module.name] = \
                        StepResult(status=StepResultStatus.SUCCESS,
                                   summary="Ran system scripts for all modules")

        console.print("Done with system steps.")

        return step_results

    def update(
            self,
            parallel=False,
            **kwargs,
    ) -> ty.Dict[str, ty.Optional[StepResult]]:
        return self.update_parallel(max_workers=None if parallel else 1, **kwargs)
        # Old school way:
        # self.for_each_module_in_dependency_chain(self.update_module)

    def update_parallel(
            self,
            max_workers: ty.Optional[int] = None,
            **kwargs,
    ) -> ty.Dict[str, ty.Optional[StepResult]]:
        from concurrent import futures
        from threading import Lock

        from armarx_setup.core.module import common

        from armarx_setup.cli.common import RichConsole
        rich_args = RichConsole._raw_kwargs()

        if max_workers is None:
            max_workers: int = axii_config.get("update_job_number")

        step_results = {}

        with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:

            # These variables are only used by the main thread.
            pending = set()

            # These variables are shared by threads.
            lock = Lock()
            module_states = dict()

            # Fetch on Axii's repo to display update notifications.
            def axii_fetch():
                from armarx_setup.self import self_update
                from armarx_setup.core.util.environ_context import EnvironContext
                from armarx_setup.core.module.update.git import git_non_interaction_env_vars
                try:
                    with EnvironContext(env=git_non_interaction_env_vars):
                        self_update.check_for_updates()
                except:
                    pass

            executor.submit(axii_fetch)

            def submit(
                    m: "Module",
                    ignore_errors=False,
                    **kwargs,
            ):
                if m.update:
                    min_result_width = 40

                    module_states[m.name] = {
                        "status": "[grey]Pending",
                        "summary": "".ljust(min_result_width),
                        "output": ""
                    }
                    pending.add(m.name)

                    def job():
                        # Wait small random amount of time to not have all jobs DOS local or server resources.
                        import time
                        import numpy as np
                        wait_time = np.random.normal(loc=1, scale=0.33)  # Wait mean 1s with stddev 0.33s.
                        if wait_time > 0:
                            time.sleep(wait_time)

                        try:
                            with lock:
                                module_states[m.name]["status"] = "[blue]Running ...[/]"
                            color = "green"
                            status = "Succeeded"

                            try:
                                step_result = m.update(capture_output=True, quiet=True, **kwargs)
                            except Exception as e:
                                step_result = common.StepResult(
                                    status=common.StepResultStatus.FAILURE,
                                    summary=f"{e}",
                                    error_output=f"Caught exception ({type(e)}) while updating "
                                                 f"module '{m.name}': {e}"
                                )

                            output = step_result.output
                            if step_result.status == common.StepResultStatus.SUCCESS_WITH_WARNING:
                                color = "yellow"
                                status = "Warning"
                            elif step_result.status == common.StepResultStatus.FAILURE:
                                color = "red"
                                status = "Failed"
                                output = step_result.error_output or step_result.output
                            elif step_result.status == common.StepResultStatus.SKIPPED:
                                color = "grey46"
                                status = "Skipped"
                            summary = step_result.summary

                            if summary is None:
                                summary = "(see output)"

                            with lock:
                                state = module_states[m.name]
                                state["status"] = f"[{color}]{status}[/]"
                                state["summary"] = summary.strip().ljust(min_result_width)
                                state["output"] = output

                            step_results[m.name] = step_result

                        except Exception as e:  # Only as a last resort.
                            console.print(f"Caught exception while updating module '{m.name}': {e}")
                            raise e

                    return executor.submit(job)

            module_to_future = self._step_with_options(
                submit, verb="update", **kwargs, sort=True,
            )

            import time
            from rich.table import Table
            from rich.live import Live
            from armarx_setup.core.util.prefix_color_map import module_to_colored_text as colorize

            def generate_table(time_passed: Optional[float] = None):
                spin_states = r"\|/–"
                update_spin_each = 0.2
                spinner = spin_states[int(time_passed / update_spin_each) % len(spin_states)] \
                    if time_passed is not None else ""
                ellipsis_states = ["...", "   ", ".  ", ".. "]
                update_ellipsis_each = 0.5
                ellipsis = ellipsis_states[
                    int(time_passed / update_ellipsis_each) % len(ellipsis_states)] \
                    if time_passed is not None else "..."
                title = "Update results" if time_passed is None else "Updating modules ..."
                table = Table("Module", f"Status {spinner}", "Summary", title=title)
                table.columns[1].width = 12
                for module, data in module_states.items():
                    table.add_row(colorize(module),
                                  data["status"].replace("...", ellipsis), data["summary"])
                return table

            # Due to a Rich issue (https://github.com/Textualize/rich/issues/2726) we need to
            # manually update the live.  refresh_per_second must be set though, so we use a very
            # small number to effectively not have any auto-refresh.
            with Live(generate_table(), refresh_per_second=1e-12, auto_refresh=False,
                      console=console) as live:
                time_passed = 0
                time_increment = 0.1

                while pending:
                    time.sleep(time_increment)
                    time_passed += time_increment

                    for module_name, future in module_to_future.items():
                        future: futures.Future

                        if module_name in pending and (future.done() or future.cancelled()):
                            # Fetch the result and free the future.
                            future.result()
                            with live.console:
                                live.console.print(f"[b]Update module[/] {colorize(module_name)} "
                                                   f"...")
                                live.console.print(module_states[module_name]["output"],
                                                   **rich_args)
                                live.console.print()

                                live.update(generate_table(time_passed=time_passed), refresh=True)
                            pending.remove(module_name)

                live.update(generate_table(), refresh=True)

        return step_results

    def update_module(self, module: "Module") -> ty.Optional[StepResult]:
        if module.update and not module.update.is_step_empty():
            console.print(f"[b]Update module '{module.name}' ...[/]")
            return module.update()

        return None

    def prepare(self, **kwargs) -> ty.Dict[str, ty.Optional[StepResult]]:
        return self._step_with_options(self.prepare_module, verb="prepare", **kwargs)

    def prepare_module(
            self,
            module: "Module",
            ignore_errors: bool = False
    ) -> ty.Optional[StepResult]:
        from armarx_setup.core.util.prefix_color_map import module_to_colored_text as colorize

        if module.prepare and not module.prepare.is_step_empty():
            console.print()
            console.print(f"[b]Prepare module[/] {colorize(module.name)} ...")

            return module.prepare(ignore_errors=ignore_errors)

        return None

    def build(self, **kwargs) -> ty.Dict[str, ty.Optional[StepResult]]:
        return self._step_with_options(self.build_module, verb="build", **kwargs)

    def build_module(
            self,
            module: "Module",
            ignore_errors: bool = False,
            clean=False
    ) -> ty.Optional[StepResult]:
        from armarx_setup.core.util.prefix_color_map import module_to_colored_text as colorize

        if module.build and not module.build.is_step_empty():
            if clean:
                return module.clean()
            else:
                console.print()
                console.print(f"[b]Build module[/] {colorize(module.name)} ...")

                return module.build(ignore_errors=ignore_errors)

        return None

    def _step_with_options(
            self,
            module_fn: ty.Callable,
            verb: str = "",
            module: str = None,
            path: str = None,
            file: str = None,
            no_deps=False,
            this=False,
            sort=False,
            **kwargs
    ) -> ty.Dict[str, ty.Any]:
        # Handle the deprecated file arg.
        if file is not None:
            if path is None:
                path = file
            else:
                raise error.AxiiError("The option 'file' is a legacy alias for 'path', so you may "
                                      "specify either one of them, but not both.")

        # We do not need this variable anymore.
        del file

        num_specified = sum([module is not None, path is not None, this])
        if num_specified > 1:
            raise error.AxiiError("You may specify either 'module', 'path' or 'this', but not more "
                                  "than one.")

        # Resolve.
        if module is not None:
            module = self.module_graph.get_module_with_name(module)
        module: Module | None

        if this:
            path = os.getcwd()

        if path is not None:
            path = os.path.abspath(os.path.expandvars(path))
            if verb:
                console.print(f"Requested to {verb} module explaining the path '{path}'.")
            module = self.module_graph.find_module_explaining_path(path, ws_path=self.path)

        def run_step(module: "Module", **kwargs):
            step_result = module_fn(module, **kwargs)
            if isinstance(step_result, StepResult):
                ignore_errors = "ignore_errors" in kwargs and kwargs["ignore_errors"]
                if step_result.status == StepResultStatus.FAILURE and not ignore_errors:
                    raise error.StepFailed(module.name, verb, step_result)
            return step_result

        if module is not None:
            if no_deps:
                chain = [module]
            else:  # Modules with their dependencies.
                chain = self.module_graph.get_dependency_chain_to_module(module)

            if sort:
                chain = sorted(chain, key=lambda m: m.name)
            return {
                module.name: run_step(module, **kwargs)
                for module in chain
            }

        else:
            if sort:
                return self.for_each_module_in_alphabetical_order(run_step, **kwargs)
            else:
                return self.for_each_module_in_dependency_chain(run_step, **kwargs)

    def _check_for_deprecated_install_module(self):
        from armarx_setup.core.module.module import Module
        from armarx_setup.core.module.common import SubStepAdapter

        def collect_executables(module: Module):
            execs = []

            if module.install is not None:
                executables_substeps = []

                if isinstance(module.install, SubStepAdapter):
                    for install in module.install.sub_steps:
                        if install.executables is not None:
                            executables_substeps.append(install.executables.execs)
                elif module.install.executables is not None:
                    executables_substeps.append(module.install.executables.execs)

                for executables_substep in executables_substeps:
                    for exec in executables_substep.keys():
                        execs.append(exec)

            return execs  # List of executables

        executables_in_workspace = self.for_each_module_in_dependency_chain(collect_executables)

        all_known_executables = []

        for module_name, known_executables in executables_in_workspace.items():
            all_known_executables += known_executables

        self._home.purge_unexplained_executables(all_known_executables)

    def install(self, **kwargs) -> ty.Dict[str, ty.Optional[StepResult]]:
        self._check_for_deprecated_install_module()
        return self._step_with_options(self.install_module, verb="install", **kwargs)

    def install_module(
            self,
            module: "Module",
            ignore_errors: bool = False
    ) -> ty.Optional[StepResult]:
        from armarx_setup.core.util.prefix_color_map import module_to_colored_text as colorize

        if module.install and not module.install.is_step_empty():
            console.print()
            console.print(f"[b]Install module[/] {colorize(module.name)} ...")
            return module.install(home=self._home, ignore_errors=ignore_errors)

        return None

    def conclude(self, **kwargs) -> ty.Dict[str, ty.Optional[StepResult]]:
        return self.for_each_module_in_dependency_chain(self.conclude_module)

    def conclude_module(self, module: "Module") -> ty.Optional[StepResult]:
        from armarx_setup.core.util.prefix_color_map import module_to_colored_text as colorize

        if module.conclude:
            console.print()
            console.print(f"[b]Conclude with module[/] {colorize(module.name)} ...")
            return module.conclude()

        return None

    def clean(self, **kwargs) -> ty.Dict[str, ty.Optional[StepResult]]:
        return self._step_with_options(lambda m, build, venv: m.clean(build=build, venv=venv), **kwargs)

    def test(self, **kwargs):
        from rich.table import Table
        from armarx_setup.core.util import prefix_color_map

        results = self.for_each_module_in_dependency_chain(self.test_module, **kwargs)

        table = Table("Module", "Result", title="Summary of Test Results")
        for module_name, result in results.items():
            if result is not None:
                if result:
                    result = "[green]Success[/]"
                else:
                    result = "[red]Failed[/]"
                table.add_row(
                    prefix_color_map.module_to_colored_text(module_name),
                    str(result),
                )
        console.print(table)

    def test_module(self, module: "Module", build: bool, **kwargs):
        from armarx_setup.core.util.prefix_color_map import module_to_colored_text as colorize

        if module.prepare is not None and not module.name.startswith("deps/"):
            if build and module.build:
                self.build_module(module)

            console.print()
            console.print(f"[b]Test module[/] {colorize(module.name)} ...")
            return module.prepare.test(**kwargs)

    def generate_env(self, **kwargs) -> ty.Dict[str, ty.Optional[StepResult]]:

        global_module: "Module" = self.config.global_
        if global_module is not None and global_module.install is not None:
            global_module.install.generate_env(self.env_file)

        self.export_module_paths(self.env_file)

        console.print(Rule(title="Setup environment from modules"))

        step_results = {}

        def gen_steps_env(module_name, steps: List["Hook"]):
            nonlocal step_results
            for step in steps:
                if step is not None:
                    step.generate_env(self.env_file)
                    step_results[module_name] = StepResult(status=StepResultStatus.SUCCESS,
                                                           summary="Successfully generated env.")

        self.for_each_module_in_dependency_chain(
            lambda m: gen_steps_env(m.name, [m.system]))
        self.for_each_module_in_dependency_chain(
            lambda m: gen_steps_env(m.name, [m.update, m.prepare, m.build, m.install]))

        self.env_file.add("PATH", "$ARMARX_WORKSPACE/.axii/bin:$PATH", by="workspace executables")

        console.print(Rule())

        self.env_file.write()

        return step_results

    def export_module_paths(self, env: "EnvFile"):
        console.print(Rule(title="Export module root paths"))

        def per_module(module: "Module"):
            if module.update:
                var_name = module.name
                var_name = var_name.replace("/", "__")
                # Replace any non-alphanumeric characters by '_'.
                var_name = re.compile(r"\W").sub("_", var_name)
                var_name = var_name + "__PATH"
                env.add(var_name, module.path, by="module graph", write=True)

        self.for_each_module_in_dependency_chain(per_module)
        console.print(Rule())
        console.print()

    def upgrade(
            self,
            env=True,
            system=True,
            update=True,
            prepare=True,
            build=True,
            install=True,
            conclude=True,
            test=False,
            ignore_errors=False,
            parallel_update=True,
            **kwargs,
    ) -> ty.Dict[str, ty.Dict[str, ty.Optional[StepResult]]]:
        from armarx_setup.core.util import cli, prefix_color_map

        def initialize_results():
            steps = ["env", "system", "update", "prepare", "build", "install", "conclude"]

            def init_step_results_for_module(module):
                step_results = {}
                for step in steps:
                    if getattr(module, step, None) is not None:  # If module has step.
                        init_step_result = StepResult(status=StepResultStatus.SKIPPED,
                                                      summary="Skipped")
                    else:
                        init_step_result = None
                    step_results[step] = init_step_result
                return step_results

            return self.for_each_module_in_dependency_chain(init_step_results_for_module)

        results: ty.Dict[str, ty.Dict[str, ty.Optional[StepResult]]] = initialize_results()

        def register_step_result(module_name: str, step: str, step_result: ty.Optional[StepResult]):
            nonlocal results

            if step_result is not None:
                if step_result.status == StepResultStatus.FAILURE and not ignore_errors:
                    raise error.StepFailed(module_name=module_name, step=step,
                                           step_result=step_result)

            results[module_name][step] = step_result

        if env:
            cli.print_full_width_heading("Setup environment")
            console.print()
            for module_name, step_result in self.generate_env().items():
                register_step_result(module_name=module_name, step="env", step_result=step_result)

        if system:
            cli.print_full_width_heading("System")
            console.print()
            for module_name, step_result in self.system(**kwargs).items():
                register_step_result(module_name=module_name, step="system",
                                     step_result=step_result)

        if update:
            cli.print_full_width_heading("Update modules")
            console.print()
            for module_name, step_result in self.update(parallel=parallel_update, **kwargs).items():
                register_step_result(module_name=module_name, step="update",
                                     step_result=step_result)

        def upgrade_module(module: "Module"):
            relevant_steps: ty.List["Hook"] = [module.prepare, module.build, module.install]
            if not any(s is not None and not s.is_step_empty() for s in relevant_steps):
                return

            color = prefix_color_map.module_to_color(module.name)
            module_text = prefix_color_map.module_to_colored_text(module.name)
            cli.print_full_width_heading(f"[b]Upgrade module[/] {module_text} ...", sep_color=color)

            if prepare:
                step_result = self.prepare_module(module, ignore_errors=ignore_errors)
                register_step_result(module_name=module.name, step="prepare",
                                     step_result=step_result)
            if build:
                step_result = self.build_module(module, ignore_errors=ignore_errors)
                register_step_result(module_name=module.name, step="build",
                                     step_result=step_result)
            if install:
                step_result = self.install_module(module, ignore_errors=ignore_errors)
                register_step_result(module_name=module.name, step="install",
                                     step_result=step_result)

        self._check_for_deprecated_install_module()
        self._step_with_options(upgrade_module, verb="upgrade", **kwargs)

        if conclude:
            cli.print_full_width_heading("Conclude")
            for module_name, step_result in self.conclude().items():
                register_step_result(module_name=module_name, step="conclude",
                                     step_result=step_result)

        if test:
            self.test(build=False)

        return results

    def auto_remove(self):
        class AutoRemoveOutput:
            def __init__(self, paths_to_remove):
                self.paths_to_remove: List[str] = paths_to_remove

            def execute(self):
                import shutil
                for path_to_remove in paths_to_remove:
                    shutil.rmtree(path_to_remove, ignore_errors=True)

        def module_filter(m):
            return os.path.join(self.path, m.name)

        paths_to_remove = []
        legit_paths = [m for m in self.for_each_module_in_dependency_chain(module_filter).values()
                       if m is not None]
        legit_paths += [os.path.join(self.path, ".axii"), os.path.join(self.path, "armarx_config")]
        for root, _, _ in os.walk(self.get_active_workspace_dir()):
            for legit_path in legit_paths:
                if root.startswith(legit_path) or root in legit_path:
                    break
            else:
                for path_to_remove in paths_to_remove:
                    if root.startswith(path_to_remove):
                        break
                else:
                    paths_to_remove.append(root)

        return AutoRemoveOutput(paths_to_remove=paths_to_remove)

    def get_deprecated_modules(self):
        class GetDeprecatedModulesOutput:
            def __init__(self, directly_added, transitive):
                self.directly_added = directly_added
                self.transitive = transitive

        deprecated_modules = [m for m in
                              self.for_each_module_in_dependency_chain(lambda m: m).values()
                              if m.general is not None and m.general.deprecated is not None]

        directly_added = []
        transitive = []
        for m in deprecated_modules:
            if m.name in self.config.modules.keys():
                directly_added.append(m)
            else:
                transitive.append(m)

        return GetDeprecatedModulesOutput(directly_added=directly_added, transitive=transitive)

    def exec(
            self,
            command: str,
            in_: Optional[str] = None,
            exclude: Optional[str] = None,
            git=False,
    ):
        from armarx_setup.core.util import commands

        rules = [
            lambda m: os.path.isdir(m.path),
        ]
        if in_:
            rules.append(lambda m: m.name.startswith(in_))
        if exclude:
            rules.append(lambda m: not m.name.startswith(exclude))
        if git:
            rules.append(lambda m: m.update is not None and m.update.git is not None)

        def exec_in_module(module: "Module"):
            if all(rule(module) for rule in rules):
                try:
                    commands.run(command, cwd=module.path)
                except error.CommandFailed as e:
                    error.handle_exception()
                    console.print(e)

        self.for_each_module_in_dependency_chain(exec_in_module)

    def notify(self, title, body, icon):
        icon_map = {
            "success": "axii_status_icon.png",
            "failure": "axii_status_icon_red.png"
        }
        icon = icon_map.get(icon, default=icon)

        icon_path = os.path.join(PACKAGE_ROOT, "resources", icon)

        try:
            import notify2  # apt install libdbus-glib-1-dev; pip install notify2 dbus-python
            notify2.init("Axii")
            notification = notify2.Notification(title, body, icon_path)
            notification.set_urgency(notify2.URGENCY_LOW)
            notification.show()
        except:
            error.handle_exception()

    def get_required_apt_packages(
            self,
            all: bool,
            missing: bool,
    ) -> ty.List[str]:
        from armarx_setup.core.workspace.system import System, SystemStep

        if not any([all, missing]):
            # Default.
            all = True

        if all and missing:
            raise error.AxiiError("Please specify only one of '--all', '--missing'.")

        modules = self.get_modules_with_system()
        system = System(ws=self)
        steps = system.plan_steps(modules=modules)
        packages_steps = filter(lambda s: s.type == SystemStep.Type.Packages, steps)

        ws_packages = set()
        for step in packages_steps:
            if all:
                step_packages = set()
                for module in step.modules:
                    assert module.system is not None, module
                    step_packages = step_packages.union(module.system.packages.apt)

            else:
                assert missing
                step_packages = step.gather_apt_packages(verbose=0)["missing"]

            ws_packages = ws_packages.union(step_packages)

        return list(ws_packages)


if __name__ == '__main__':
    from armarx_setup.core.module import Module
    from armarx_setup.core.module.common import Hook
    from armarx_setup.core.module_graph import ModuleGraph
    from armarx_setup.core.workspace.config import Config
    from armarx_setup.core.workspace.env import EnvFile
