#!/usr/bin/env bash

# General comments
###################################################################

## > Extensions

# A set of extensions will be installed. Eventually, they are configured through 
# the provided 'settings.json' file.

## > User config

# Default settings are installed, including keybindings which are heavily 
# inspired by QtCreator.

# Note: If you manually execute this file, e.g. through 'code_configure', the 
#       setting files will be overriden. 

## > Executable `_code` vs `code`:

# The `_code` executable is provided by the vs_code/{version} module.
# This module overrides the `code` executable to provide a pre-exec hook. Therefore,
# using the `code` executable here would result in infinite recursion.


# Preparation
###################################################################

SCRIPT_DIR=$(cd -- "$(dirname -- "${BASH_SOURCE[0]}")" &>/dev/null && pwd)

source $SCRIPT_DIR/_colors.sh

# Helper functions
###################################################################

function _print_colored
{
  Text=$1
  Color=$2

  # enables the colored output only for the given text
  echo -e "${Color}${Text}${Color_Off}"
}

# we only replace the enable-listed environment variables
function _configure_file {

  source_file="$1"
  target_file="$2"

  # echo "Resolve environment variables and install file '$target_file' from '$source_file"
  envsubst '$VS_CODE_USER_NAME,$VS_CODE_USER_EMAIL' <"$source_file" >"$target_file"
}

function _install_file {
  cp $1 $2
}

# Functions to configure VS code
###################################################################

function _code_install_extensions {

  extensions=(
    "jeff-hykin.better-cpp-syntax"
    "llvm-vs-code-extensions.vscode-clangd"
    "anjali.clipboard-history"
    "twxs.cmake"
    "bbenoist.doxygen"
    "cschlosser.doxdocgen"
    "ms-vscode-remote.vscode-remote-extensionpack"
    "zerocinc.slice"
    "gruntfuggly.todo-tree"
    "dotjoshjohnson.xml"
    "redhat.vscode-yaml"
    "ms-python.python"
    "ms-python.vscode-pylance"
    "liviuschera.noctis"
    "ms-toolsai.jupyter"
    "mads-hartmann.bash-ide-vscode"
    "zerocinc.slice"
    "ms-vscode-remote.remote-ssh"
    "ms-vscode-remote.remote-ssh-edit"
    "ms-vscode.remote-explorer"
    "waderyan.gitblame"
    "eamodio.gitlens"
    "foxundermoon.shell-format"
    "amiralizadeh9480.cpp-helper"
    "redhat.vscode-xml"
  )

  _print_colored "> Installing extensions ... \n" $BGreen

  for extension in ${extensions[@]}; do
    _print_colored "Installing extension $extension ..." $Green
    _code --install-extension $extension --force
    printf "\n"
  done

}

# Install user data into the workspace .axii/config folder
###################################################################

function _code_install_user_data {

  _print_colored "> Configuring settings, keybindings and tasks ...\n" $BGreen

  SCRIPT_DIR=$(cd -- "$(dirname -- "${BASH_SOURCE[0]}")" &>/dev/null && pwd)

  config_dir="$SCRIPT_DIR"
  user_data_config_dir="$SCRIPT_DIR/user_data"

  # the files to be copied / symlinked
  # Note: we chose the file extension .json2 to avoid that axii treats it as a module
  keybindings_file="$user_data_config_dir/User/keybindings.json2"
  settings_file="$user_data_config_dir/User/settings.json2"
  tasks_file="$user_data_config_dir/User/tasks.json2"

  snippets_dir="$user_data_config_dir/User/snippets/"

  snippet_files=(
    "$snippets_dir/armarx-cmake.code-snippets"
    "$snippets_dir/armarx-header.code-snippets"
    "$snippets_dir/armarx-source.code-snippets"
    "$snippets_dir/armarx-license-header.code-snippets"
    "$snippets_dir/simox-license-header.code-snippets"
  )

  user_data_dir="$ARMARX_WORKSPACE/.axii/config/vscode/user_data/User"

  _print_colored "Creating directory '$user_data_dir'" $Green
  mkdir -p $user_data_dir

  printf "\n"

  _print_colored "Installing file: " $Green
  _print_colored " - keybindings.json" $Green
  _install_file "$keybindings_file" "$user_data_dir/keybindings.json"

  _print_colored " - settings.json" $Green
  _install_file "$settings_file" "$user_data_dir/settings.json"

  _print_colored " - tasks.json" $Green
  _install_file "$tasks_file" "$user_data_dir/tasks.json"

  printf "\n"
  _print_colored "> Configuring and installing snippets ...\n" $BGreen

  # _print_colored "Creating directory $snippets_target_dir" $Green
  snippets_target_dir=$user_data_dir/snippets
  mkdir -p $snippets_target_dir

  VS_CODE_USER_NAME=$(git config --get user.name)
  VS_CODE_USER_EMAIL_RAW=$(git config --get user.email)

  _print_colored "Configuring snippets for Git user '$VS_CODE_USER_NAME' with e-mail '$VS_CODE_USER_EMAIL_RAW'" $Green

  # change 'foo.bar@baz.de' to 'foo dot bar at baz dot de'
  VS_CODE_USER_EMAIL=$(echo $VS_CODE_USER_EMAIL_RAW | sed 's/[.]/ dot /g' | sed 's/[@]/ at /g')

  function _install_snippet {
    snippet_source_filename=$1
    snippet_target_filename=axii-$(basename ${snippet_source_filename})

    _print_colored " - $snippet_target_filename" $Green
    _configure_file "$snippet_source_filename" "$snippets_target_dir/$snippet_target_filename"
  }

  printf "\n"

  # configure and install snippets into directory
  _print_colored "Installing snippets:" $Green
  for snippet_file in ${snippet_files[@]}; do
    _install_snippet $snippet_file
  done

}

_print_colored "\nConfiguring VS code\n" $BBlue

_code_install_extensions
_code_install_user_data
