"""Unit tests for the armarx_setup.core.util.commands module."""

import pytest

from armarx_setup.core import error
from armarx_setup.core.util.commands import execute, StdInBuffer


def test_run_success():
    out = execute("echo this is a test".split(), cwd="/")
    assert out == "this is a test"


def test_run_success_callback():
    streamed_out = ""

    def stdout_cb(text):
        nonlocal streamed_out
        streamed_out += text

    out = execute("echo this is a test".split(), cwd="/", stdout_callback=stdout_cb)

    assert out == "this is a test"
    assert streamed_out == "this is a test"


def test_run_exception_and_callback():
    streamed_out = ""

    def stdout_cb(text):
        nonlocal streamed_out
        streamed_out += text

    with pytest.raises(error.CommandFailed) as e_info:
        execute(["bash", "-c", "echo testing && false"], cwd="/", stdout_callback=stdout_cb)

    assert e_info.value.cmd == ["bash", "-c", "echo testing && false"]
    assert e_info.value.cwd == "/"
    assert e_info.value.return_code == 1

    assert streamed_out == "testing"


def test_run_stdin():
    pipe = execute(["echo", "test1\ntest2\ntest3"], cwd="/")
    assert pipe == "test1\ntest2\ntest3"
    pipe = pipe + "\n\x04"  # xargs requires EOF on a new line to terminate.

    out = execute("xargs echo".split(), cwd="/", stdin=pipe, stdout_callback=print)
    assert out == "test1 test2 test3"


def test_async_stdin():
    stdin = StdInBuffer()
    you_said = ""

    def stdout_cb(text):
        nonlocal you_said
        if "Yes or no?" in text:
            stdin.write_line("yes")
        elif "You said" in text:
            you_said = text

    query = ["bash", "-c", 'echo "Yes or no?" && read decision && echo "You said -$decision-"']
    out = execute(query, cwd="/", stdin=stdin, stdout_callback=stdout_cb)

    assert "Yes or no?" in out
    assert "You said" in out

    assert you_said == "You said -yes-"


def test_shell_lex():
    out1 = execute("echo 'this is test 1'")
    out2 = execute('echo "this is test 2"')
    out3 = execute('echo "\\"Yes, absolutely!\\"", he said')

    assert out1 == "this is test 1"
    assert out2 == "this is test 2"
    assert out3 == '"Yes, absolutely!", he said'
