import os
import shutil

import pytest

from armarx_setup.core.workspace.home import Home

exec_config = {
        "exec": "te",
        "exec_args": ["--test-arg", "--test-arg2"],
        "pre_exec_commands": ["echo 'pre_exec_command'"],
        "post_exec_commands": ["echo 'post_exec_command'"]
    }

test_workspace_name = "test_workspace"
testing_directory = "/tmp"
workspace_path = os.path.join(testing_directory, test_workspace_name)


@pytest.fixture
def setup():
    os.mkdir(workspace_path)
    home = Home(workspace_path)

    yield home, workspace_path

    shutil.rmtree(workspace_path)


def test_home():
    home = Home(workspace_path)
    assert home.home_path == os.path.join(workspace_path, ".axii")
    assert home.bin_path == os.path.join(home.home_path, "bin")


def check_strings_in_content(strings, content):
    test_name = "write executable"
    # Split the content string into lines
    content_lines = content.split('\n')

    # Check if every string in the list is in the content lines
    for string in strings:
        assert string in content_lines, f"'{string}' not found in content at test {test_name}"


def test_write_executable(setup):
    home, _ = setup

    name = "test_exec"
    kill_trap = False
    fork = ""
    home.install_executable(name, exec_config, kill_trap, fork)

    with open(os.path.join(home.bin_path, name)) as f:
        content = f.read()

    kill_trap_assert_error = f"'trap : SIGTERM SIGINT' found in content at test write executable"
    assert not ("trap : SIGTERM SIGINT" in content), kill_trap_assert_error
    check_strings_in_content(exec_config["pre_exec_commands"], content)
    check_strings_in_content(exec_config["post_exec_commands"], content)

    exec_args = " ".join(exec_config["exec_args"])
    exec_string = f"{exec_config['exec']} {exec_args} $@ \n"
    assert_error_message = f"'{exec_string}' not found in content at test write executable"
    assert exec_string in content, assert_error_message

    def check_order(main_str, substrings):
        indices = [main_str.index(s) if s in main_str else -1 for s in substrings]
        return all(i < j for i, j in zip(indices, indices[1:]))

    assert check_order(content, exec_config["pre_exec_commands"]), "Pre-exec commands not in correct order"
    assert check_order(content, exec_config["post_exec_commands"]), "Post-exec commands not in correct order"
    # check for order in test_exec: binbash, pre_exec_commands, exec, post_exec_commands
    order = ["#!/bin/bash",
             "\n".join(exec_config["pre_exec_commands"]),
             exec_string,
             "\n".join(exec_config["post_exec_commands"])]
    assert check_order(content, order), "Content not in correct order"


def test_write_executable_with_kill_trap(setup):
    home, _ = setup

    name = "test_exec"
    kill_trap = True
    fork = "&"
    home.install_executable(name, exec_config, kill_trap, fork)

    with open(os.path.join(home.bin_path, name)) as f:
        content = f.read()

    kill_trap_assert_error = f"'trap : SIGTERM SIGINT' not found in content at test write executable"
    assert "trap : SIGTERM SIGINT" in content, kill_trap_assert_error
    check_strings_in_content(exec_config["pre_exec_commands"], content)
    check_strings_in_content(exec_config["post_exec_commands"], content)

    exec_args = " ".join(exec_config["exec_args"])
    exec_string = f"{exec_config['exec']} {exec_args} $@ {fork}\n"
    assert_error_message = f"'{exec_string}' not found in content at test write executable"
    assert exec_string in content, assert_error_message

    def check_order(main_str, substrings):
        indices = [main_str.index(s) if s in main_str else -1 for s in substrings]
        return all(i < j for i, j in zip(indices, indices[1:]))

    assert check_order(content, exec_config["pre_exec_commands"]), "Pre-exec commands not in correct order"
    assert check_order(content, exec_config["post_exec_commands"]), "Post-exec commands not in correct order"
    order = ["#!/bin/bash",
             "\n".join(exec_config["pre_exec_commands"]),
             exec_string,
             "\n".join(exec_config["post_exec_commands"])]
    assert check_order(content, order), "Content not in correct order"
